# Copyright (c) 2016 Ansible by Red Hat, Inc.
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import json
import sys
import threading
import logging

from ansible_runner import output
from ansible_runner.config.runner import RunnerConfig
from ansible_runner.config.command import CommandConfig
from ansible_runner.config.inventory import InventoryConfig
from ansible_runner.config.ansible_cfg import AnsibleCfgConfig
from ansible_runner.config.doc import DocConfig
from ansible_runner.runner import Runner
from ansible_runner.streaming import Transmitter, Worker, Processor
from ansible_runner.utils import (
    dump_artifacts,
    check_isolation_executable_installed,
    sanitize_json_response,
    signal_handler,
)

logging.getLogger('ansible-runner').addHandler(logging.NullHandler())


def init_runner(**kwargs):
    '''
    Initialize the Runner() instance

    This function will properly initialize both run() and run_async()
    functions in the same way and return a value instance of Runner.

    See parameters given to :py:func:`ansible_runner.interface.run`
    '''

    # Handle logging first thing
    debug = kwargs.pop('debug', None)
    logfile = kwargs.pop('logfile', None)

    if not kwargs.pop("ignore_logging", True):
        output.configure()
        if debug in (True, False):
            output.set_debug('enable' if debug is True else 'disable')

        if logfile:
            output.set_logfile(logfile)

    # If running via the transmit-worker-process method, we must only extract things as read-only
    # inside of one of these commands. That could be either transmit or worker.
    if kwargs.get('streamer') not in ('worker', 'process'):
        dump_artifacts(kwargs)

    if kwargs.get('streamer'):
        # undo any full paths that were dumped by dump_artifacts above in the streamer case
        private_data_dir = kwargs['private_data_dir']
        project_dir = os.path.join(private_data_dir, 'project')

        playbook_path = kwargs.get('playbook') or ''
        if os.path.isabs(playbook_path) and playbook_path.startswith(project_dir):
            kwargs['playbook'] = os.path.relpath(playbook_path, project_dir)

        inventory_path = kwargs.get('inventory') or ''
        if os.path.isabs(inventory_path) and inventory_path.startswith(private_data_dir):
            kwargs['inventory'] = os.path.relpath(inventory_path, private_data_dir)

        roles_path = kwargs.get('envvars', {}).get('ANSIBLE_ROLES_PATH') or ''
        if os.path.isabs(roles_path) and roles_path.startswith(private_data_dir):
            kwargs['envvars']['ANSIBLE_ROLES_PATH'] = os.path.relpath(roles_path, private_data_dir)

    event_callback_handler = kwargs.pop('event_handler', None)
    status_callback_handler = kwargs.pop('status_handler', None)
    artifacts_handler = kwargs.pop('artifacts_handler', None)
    cancel_callback = kwargs.pop('cancel_callback', None)
    if cancel_callback is None:
        # attempt to load signal handler.
        # will return None if we are not in the main thread
        cancel_callback = signal_handler()
    finished_callback = kwargs.pop('finished_callback', None)

    streamer = kwargs.pop('streamer', None)
    if streamer:
        if streamer == 'transmit':
            stream_transmitter = Transmitter(**kwargs)
            return stream_transmitter

        if streamer == 'worker':
            stream_worker = Worker(**kwargs)
            return stream_worker

        if streamer == 'process':
            stream_processor = Processor(event_handler=event_callback_handler,
                                         status_handler=status_callback_handler,
                                         artifacts_handler=artifacts_handler,
                                         cancel_callback=cancel_callback,
                                         finished_callback=finished_callback,
                                         **kwargs)
            return stream_processor

    if kwargs.get("process_isolation", False):
        pi_executable = kwargs.get("process_isolation_executable", "podman")
        if not check_isolation_executable_installed(pi_executable):
            print(f'Unable to find process isolation executable: {pi_executable}')
            sys.exit(1)

    kwargs.pop('_input', None)
    kwargs.pop('_output', None)
    rc = RunnerConfig(**kwargs)
    rc.prepare()

    return Runner(rc,
                  event_handler=event_callback_handler,
                  status_handler=status_callback_handler,
                  artifacts_handler=artifacts_handler,
                  cancel_callback=cancel_callback,
                  finished_callback=finished_callback)


def run(**kwargs):
    '''
    Run an Ansible Runner task in the foreground and return a Runner object when complete.

    :param str private_data_dir: The directory containing all runner metadata needed to invoke the runner
                             module. Output artifacts will also be stored here for later consumption.
    :param str ident: The run identifier for this invocation of Runner. Will be used to create and name
                  the artifact directory holding the results of the invocation.
    :param bool json_mode: Store event data in place of stdout on the console and in the stdout file
    :param str or list playbook: The playbook (either a list or dictionary of plays, or as a path relative to
                     ``private_data_dir/project``) that will be invoked by runner when executing Ansible.
    :param str module: The module that will be invoked in ad-hoc mode by runner when executing Ansible.
    :param str module_args: The module arguments that will be supplied to ad-hoc mode.
    :param str host_pattern: The host pattern to match when running in ad-hoc mode.
    :param str or dict or list inventory: Overrides the inventory directory/file (supplied at ``private_data_dir/inventory``) with
        a specific host or list of hosts. This can take the form of:

            - Path to the inventory file in the ``private_data_dir/inventory`` directory or
              an absolute path to the inventory file
            - Native python dict supporting the YAML/json inventory structure
            - A text INI formatted string
            - A list of inventory sources, or an empty list to disable passing inventory

    :param str role: Name of the role to execute.
    :param str or list roles_path: Directory or list of directories to assign to ANSIBLE_ROLES_PATH
    :param dict envvars: Environment variables to be used when running Ansible. Environment variables will also be
                    read from ``env/envvars`` in ``private_data_dir``
    :param dict extravars: Extra variables to be passed to Ansible at runtime using ``-e``. Extra vars will also be
                      read from ``env/extravars`` in ``private_data_dir``.
    :param dict passwords: A dictionary containing password prompt patterns and response values used when processing output from
                      Ansible. Passwords will also be read from ``env/passwords`` in ``private_data_dir``.
    :param dict settings: A dictionary containing settings values for the ``ansible-runner`` runtime environment. These will also
                     be read from ``env/settings`` in ``private_data_dir``.
    :param str ssh_key: The ssh private key passed to ``ssh-agent`` as part of the ansible-playbook run.
    :param str cmdline: Command line options passed to Ansible read from ``env/cmdline`` in ``private_data_dir``
    :param bool suppress_env_files: Disable the writing of files into the ``env`` which may store sensitive information
    :param str limit: Matches ansible's ``--limit`` parameter to further constrain the inventory to be used
    :param int forks: Control Ansible parallel concurrency
    :param int verbosity: Control how verbose the output of ansible-playbook is
    :param bool quiet: Disable all output
    :param str artifact_dir: The path to the directory where artifacts should live, this defaults to 'artifacts' under the private data dir
    :param str project_dir: The path to the playbook content, this defaults to 'project' within the private data dir
    :param int rotate_artifacts: Keep at most n artifact directories, disable with a value of 0 which is the default
    :param int timeout: The timeout value in seconds that will be passed to either ``pexpect`` of ``subprocess`` invocation
                    (based on ``runner_mode`` selected) while executing command. It the timeout is triggered it will force cancel the
                    execution.
    :param str streamer: Optionally invoke ansible-runner as one of the steps in the streaming pipeline
    :param io.FileIO _input: An optional file or file-like object for use as input in a streaming pipeline
    :param io.FileIO _output: An optional file or file-like object for use as output in a streaming pipeline
    :param Callable event_handler: An optional callback that will be invoked any time an event is received by Runner itself, return True to keep the event
    :param Callable cancel_callback: An optional callback that can inform runner to cancel (returning True) or not (returning False)
    :param Callable finished_callback: An optional callback that will be invoked at shutdown after process cleanup.
    :param Callable status_handler: An optional callback that will be invoked any time the status changes (e.g...started, running, failed, successful, timeout)
    :param Callable artifacts_handler: An optional callback that will be invoked at the end of the run to deal with the artifacts from the run.
    :param bool process_isolation: Enable process isolation, using either a container engine (e.g. podman) or a sandbox (e.g. bwrap).
    :param str process_isolation_executable: Process isolation executable or container engine used to isolate execution. (default: podman)
    :param str process_isolation_path: Path that an isolated playbook run will use for staging. (default: /tmp)
    :param str or list process_isolation_hide_paths: A path or list of paths on the system that should be hidden from the playbook run.
    :param str or list process_isolation_show_paths: A path or list of paths on the system that should be exposed to the playbook run.
    :param str or list process_isolation_ro_paths: A path or list of paths on the system that should be exposed to the playbook run as read-only.
    :param str container_image: Container image to use when running an ansible task
    :param list container_volume_mounts: List of bind mounts in the form 'host_dir:/container_dir. (default: None)
    :param list container_options: List of container options to pass to execution engine.
    :param str directory_isolation_base_path: An optional path will be used as the base path to create a temp directory, the project contents will be
                                          copied to this location which will then be used as the working directory during playbook execution.
    :param str fact_cache: A string that will be used as the name for the subdirectory of the fact cache in artifacts directory.
                       This is only used for 'jsonfile' type fact caches.
    :param str fact_cache_type: A string of the type of fact cache to use.  Defaults to 'jsonfile'.
    :param bool omit_event_data: Omits extra ansible event data from event payload (stdout and event still included)
    :param bool only_failed_event_data: Omits extra ansible event data unless it's a failed event (stdout and event still included)
    :param bool check_job_event_data: Check if job events data is completely generated. If event data is not completely generated and if
                                 value is set to 'True' it will raise 'AnsibleRunnerException' exception,
                                 if set to 'False' it log a debug message and continue execution. Default value is 'False'

    :returns: A :py:class:`ansible_runner.runner.Runner` object, or a simple object containing ``rc`` if run remotely
    '''
    r = init_runner(**kwargs)
    r.run()
    return r


def run_async(**kwargs):
    '''
    Runs an Ansible Runner task in the background which will start immediately. Returns the thread object and a Runner object.

    This uses the same parameters as :py:func:`ansible_runner.interface.run`

    :returns: A tuple containing a :py:class:`threading.Thread` object and a :py:class:`ansible_runner.runner.Runner` object
    '''
    r = init_runner(**kwargs)
    runner_thread = threading.Thread(target=r.run)
    runner_thread.start()
    return runner_thread, r


def init_command_config(executable_cmd, cmdline_args=None, **kwargs):
    '''
    Initialize the Runner() instance

    This function will properly initialize both run_command() and run_command_async()
    functions in the same way and return a value instance of Runner.

    See parameters given to :py:func:`ansible_runner.interface.run_command`
    '''
    event_callback_handler = kwargs.pop('event_handler', None)
    status_callback_handler = kwargs.pop('status_handler', None)
    artifacts_handler = kwargs.pop('artifacts_handler', None)
    cancel_callback = kwargs.pop('cancel_callback', None)
    finished_callback = kwargs.pop('finished_callback', None)

    rc = CommandConfig(**kwargs)
    rc.prepare_run_command(executable_cmd, cmdline_args=cmdline_args)
    return Runner(rc,
                  event_handler=event_callback_handler,
                  status_handler=status_callback_handler,
                  artifacts_handler=artifacts_handler,
                  cancel_callback=cancel_callback,
                  finished_callback=finished_callback)


def run_command(executable_cmd, cmdline_args=None, **kwargs):
    '''
    Run an (Ansible) commands in the foreground and return a Runner object when complete.

    :param str executable_cmd: The command to be executed.
    :param list cmdline_args: A list of arguments to be passed to the executable command.
    :param int input_fd: This parameter is applicable when ``runner_mode`` is set to ``subprocess``, it provides the
                     input file descrption to interact with the sub-process running the command.
    :param int output_fd: The output file descriptor to stream the output of command execution.
    :param int error_fd: This parameter is applicable when ``runner_mode`` is set to ``subprocess``, it provides the
                     error file descrption to read the error received while executing the command.
    :param str runner_mode: The applicable values are ``pexpect`` and ``subprocess``. If the value of ``input_fd`` parameter
                        is set or the executable command is one of ``ansible-config``, ``ansible-doc`` or ``ansible-galaxy``
                        the default value is set to ``subprocess`` else in other cases it is set to ``pexpect``.
    :param str host_cwd: The host current working directory to be mounted within the container (if enabled) and will be
                     the work directory within container.
    :param dict envvars: Environment variables to be used when running Ansible. Environment variables will also be
                    read from ``env/envvars`` in ``private_data_dir``
    :param dict passwords: A dictionary containing password prompt patterns and response values used when processing output from
                      Ansible. Passwords will also be read from ``env/passwords`` in ``private_data_dir``.
    :param dict settings: A dictionary containing settings values for the ``ansible-runner`` runtime environment. These will also
                     be read from ``env/settings`` in ``private_data_dir``.
    :param str ssh_key: The ssh private key passed to ``ssh-agent`` as part of the ansible-playbook run.
    :param bool quiet: Disable all output
    :param bool json_mode: Store event data in place of stdout on the console and in the stdout file
    :param str artifact_dir: The path to the directory where artifacts should live, this defaults to 'artifacts' under the private data dir
    :param str project_dir: The path to the playbook content, this defaults to 'project' within the private data dir
    :param int rotate_artifacts: Keep at most n artifact directories, disable with a value of 0 which is the default
    :param int timeout: The timeout value in seconds that will be passed to either ``pexpect`` of ``subprocess`` invocation
                    (based on ``runner_mode`` selected) while executing command. It the timeout is triggered it will force cancel the
                    execution.
    :param bool process_isolation: Enable process isolation, using a container engine (e.g. podman).
    :param str process_isolation_executable: Process isolation executable or container engine used to isolate execution. (default: podman)
    :param str container_image: Container image to use when running an ansible task
    :param list container_volume_mounts: List of bind mounts in the form 'host_dir:/container_dir:labels. (default: None)
    :param list container_options: List of container options to pass to execution engine.
    :param str container_workdir: The working directory within the container.
    :param str fact_cache: A string that will be used as the name for the subdirectory of the fact cache in artifacts directory.
                       This is only used for 'jsonfile' type fact caches.
    :param str fact_cache_type: A string of the type of fact cache to use.  Defaults to 'jsonfile'.
    :param str private_data_dir: The directory containing all runner metadata needed to invoke the runner
                             module. Output artifacts will also be stored here for later consumption.
    :param str ident: The run identifier for this invocation of Runner. Will be used to create and name
                  the artifact directory holding the results of the invocation.
    :param Callable event_handler: An optional callback that will be invoked any time an event is received by Runner itself, return True to keep the event
    :param Callable cancel_callback: An optional callback that can inform runner to cancel (returning True) or not (returning False)
    :param Callable finished_callback: An optional callback that will be invoked at shutdown after process cleanup.
    :param Callable status_handler: An optional callback that will be invoked any time the status changes (e.g...started, running, failed, successful, timeout)
    :param Callable artifacts_handler: An optional callback that will be invoked at the end of the run to deal with the artifacts from the run.
    :param bool check_job_event_data: Check if job events data is completely generated. If event data is not completely generated and if
                                 value is set to 'True' it will raise 'AnsibleRunnerException' exception,
                                 if set to 'False' it log a debug message and continue execution. Default value is 'False'

    :returns: Returns a tuple of response, error string and return code.
              In case if ``runner_mode`` is set to ``pexpect`` the error value is empty as
              ``pexpect`` uses same output descriptor for stdout and stderr.
    '''
    r = init_command_config(executable_cmd, cmdline_args=cmdline_args, **kwargs)
    r.run()
    with r.stdout as stdout, r.stderr as stderr:
        response = stdout.read()
        error = stderr.read()
    return response, error, r.rc


def run_command_async(executable_cmd, cmdline_args=None, **kwargs):
    '''
    Run an (Ansible) commands in the background which will start immediately. Returns the thread object and a Runner object.

    This uses the same parameters as :py:func:`ansible_runner.interface.run_command`

    :returns: A tuple containing a :py:class:`threading.Thread` object and a :py:class:`ansible_runner.runner.Runner` object
    '''
    r = init_command_config(executable_cmd, cmdline_args=cmdline_args, **kwargs)
    runner_thread = threading.Thread(target=r.run)
    runner_thread.start()
    return runner_thread, r


def init_plugin_docs_config(plugin_names, plugin_type=None, response_format=None,
                            snippet=False, playbook_dir=None, module_path=None, **kwargs):
    '''
    Initialize the Runner() instance

    This function will properly initialize both get_plugin_docs() and get_plugin_docs_async()
    functions in the same way and return a value instance of Runner.

    See parameters given to :py:func:`ansible_runner.interface.get_plugin_docs`
    '''

    event_callback_handler = kwargs.pop('event_handler', None)
    status_callback_handler = kwargs.pop('status_handler', None)
    artifacts_handler = kwargs.pop('artifacts_handler', None)
    cancel_callback = kwargs.pop('cancel_callback', None)
    finished_callback = kwargs.pop('finished_callback', None)

    rd = DocConfig(**kwargs)
    rd.prepare_plugin_docs_command(plugin_names, plugin_type=plugin_type, response_format=response_format,
                                   snippet=snippet, playbook_dir=playbook_dir, module_path=module_path)
    return Runner(rd, event_handler=event_callback_handler, status_handler=status_callback_handler, artifacts_handler=artifacts_handler,
                  cancel_callback=cancel_callback, finished_callback=finished_callback)


def get_plugin_docs(plugin_names, plugin_type=None, response_format=None, snippet=False, playbook_dir=None, module_path=None, **kwargs):
    '''
    Run an ansible-doc command to get plugin docs  in the foreground and return a Runner object when complete.

    :param plugin_names: The name of the plugins to get docs.
    :param plugin_type: The type of the plugin mentioned in plugins_names. Valid values are ``become``, ``cache``, ``callback``,
                        ``cliconf``, ``connection``, ``httpapi``, ``inventory``, ``lookup``, ``netconf``, ``shell``, ``vars``,
                        ``module``, ``strategy``. If the value is not provided it defaults to ``module``.
    :param response_format: The output format for response. Valid values can be one of ``json`` or ``human`` and the response
                            is either json string or plain text in human readable foramt. Default value is ``json``.
    :param snippet: Show playbook snippet for specified plugin(s).
    :param playbook_dir: This parameter is used to sets the relative path to handle playbook adjacent installed plugins.
    :param module_path: This parameter is prepend colon-separated path(s) to module library
                        (default=~/.ansible/plugins/modules:/usr/share/ansible/plugins/modules).
    :param runner_mode: The applicable values are ``pexpect`` and ``subprocess``. Default is set to ``subprocess``.
    :param host_cwd: The host current working directory to be mounted within the container (if enabled) and will be
                     the work directory within container.
    :param envvars: Environment variables to be used when running Ansible. Environment variables will also be
                    read from ``env/envvars`` in ``private_data_dir``
    :param passwords: A dictionary containing password prompt patterns and response values used when processing output from
                      Ansible. Passwords will also be read from ``env/passwords`` in ``private_data_dir``.
    :param settings: A dictionary containing settings values for the ``ansible-runner`` runtime environment. These will also
                     be read from ``env/settings`` in ``private_data_dir``.
    :param ssh_key: The ssh private key passed to ``ssh-agent`` as part of the ansible-playbook run.
    :param quiet: Disable all output
    :param json_mode: Store event data in place of stdout on the console and in the stdout file
    :param artifact_dir: The path to the directory where artifacts should live, this defaults to 'artifacts' under the private data dir
    :param project_dir: The path to the playbook content, this defaults to 'project' within the private data dir
    :param rotate_artifacts: Keep at most n artifact directories, disable with a value of 0 which is the default
    :param timeout: The timeout value in seconds that will be passed to either ``pexpect`` of ``subprocess`` invocation
                    (based on ``runner_mode`` selected) while executing command. It the timeout is triggered it will force cancel the
                    execution.
    :param process_isolation: Enable process isolation, using a container engine (e.g. podman).
    :param process_isolation_executable: Process isolation executable or container engine used to isolate execution. (default: podman)
    :param container_image: Container image to use when running an ansible task
    :param container_volume_mounts: List of bind mounts in the form 'host_dir:/container_dir:labels. (default: None)
    :param container_options: List of container options to pass to execution engine.
    :param container_workdir: The working directory within the container.
    :param fact_cache: A string that will be used as the name for the subdirectory of the fact cache in artifacts directory.
                       This is only used for 'jsonfile' type fact caches.
    :param fact_cache_type: A string of the type of fact cache to use.  Defaults to 'jsonfile'.
    :param private_data_dir: The directory containing all runner metadata needed to invoke the runner
                             module. Output artifacts will also be stored here for later consumption.
    :param ident: The run identifier for this invocation of Runner. Will be used to create and name
                  the artifact directory holding the results of the invocation.
    :param event_handler: An optional callback that will be invoked any time an event is received by Runner itself, return True to keep the event
    :param cancel_callback: An optional callback that can inform runner to cancel (returning True) or not (returning False)
    :param finished_callback: An optional callback that will be invoked at shutdown after process cleanup.
    :param status_handler: An optional callback that will be invoked any time the status changes (e.g...started, running, failed, successful, timeout)
    :param artifacts_handler: An optional callback that will be invoked at the end of the run to deal with the artifacts from the run.
    :param check_job_event_data: Check if job events data is completely generated. If event data is not completely generated and if
                                 value is set to 'True' it will raise 'AnsibleRunnerException' exception,
                                 if set to 'False' it log a debug message and continue execution. Default value is 'False'

    :type plugin_names: list
    :type plugin_type: str
    :type response_format: str
    :type snippet: bool
    :type playbook_dir: str
    :type module_path: str
    :type runner_mode: str
    :type host_cwd: str
    :type envvars: dict
    :type passwords: dict
    :type settings: dict
    :type private_data_dir: str
    :type project_dir: str
    :type artifact_dir: str
    :type fact_cache_type: str
    :type fact_cache: str
    :type process_isolation: bool
    :type process_isolation_executable: str
    :type container_image: str
    :type container_volume_mounts: list
    :type container_options: list
    :type container_workdir: str
    :type ident: str
    :type rotate_artifacts: int
    :type timeout: int
    :type ssh_key: str
    :type quiet: bool
    :type json_mode: bool
    :type event_handler: Callable
    :type cancel_callback: Callable
    :type finished_callback: Callable
    :type status_handler: Callable
    :type artifacts_handler: Callable
    :type check_job_event_data: bool

    :returns: Returns a tuple of response and error string. In case if ``runner_mode`` is set to ``pexpect`` the error value is empty as
              ``pexpect`` uses same output descriptor for stdout and stderr. If the value of ``response_format`` is ``json``
              it returns a python dictionary object.
    '''
    r = init_plugin_docs_config(plugin_names, plugin_type=plugin_type, response_format=response_format,
                                snippet=snippet, playbook_dir=playbook_dir, module_path=module_path, **kwargs)
    r.run()
    with r.stdout as stdout, r.stderr as stderr:
        response = stdout.read()
        error = stderr.read()
    if response and response_format == 'json':
        response = json.loads(sanitize_json_response(response))
    return response, error


def get_plugin_docs_async(plugin_names, plugin_type=None, response_format=None, snippet=False, playbook_dir=None, module_path=None, **kwargs):
    '''
    Run an ansible-doc command in the background which will start immediately. Returns the thread object and a Runner object.

    This uses the same parameters as :py:func:`ansible_runner.interface.get_plugin_docs`

    :returns: A tuple containing a :py:class:`threading.Thread` object and a :py:class:`ansible_runner.runner.Runner` object
    '''
    r = init_plugin_docs_config(plugin_names, plugin_type=plugin_type, response_format=response_format,
                                snippet=snippet, playbook_dir=playbook_dir, module_path=module_path, **kwargs)
    doc_runner_thread = threading.Thread(target=r.run)
    doc_runner_thread.start()
    return doc_runner_thread, r


def get_plugin_list(list_files=None, response_format=None, plugin_type=None, playbook_dir=None, module_path=None, **kwargs):
    '''
    Run an ansible-doc command to get list of installed Ansible plugins.

    :param list_files: The boolean parameter is set to ``True`` returns file path of the plugin along with the plugin name.
    :param response_format: The output format for response. Valid values can be one of ``json`` or ``human`` and the response
                            is either json string or plain text in human readable foramt. Default value is ``json``.
    :param plugin_type: The type of the plugin mentioned in plugins_names. Valid values are ``become``, ``cache``, ``callback``,
                        ``cliconf``, ``connection``, ``httpapi``, ``inventory``, ``lookup``, ``netconf``, ``shell``, ``vars``,
                        ``module``, ``strategy``. If the value is not provided it defaults to ``module``.
    :param playbook_dir: This parameter is used to sets the relative path to handle playbook adjacent installed plugins.
    :param module_path: This parameter is prepend colon-separated path(s) to module library
                        (default=~/.ansible/plugins/modules:/usr/share/ansible/plugins/modules).
    :param runner_mode: The applicable values are ``pexpect`` and ``subprocess``. Default is set to ``subprocess``.
    :param host_cwd: The host current working directory to be mounted within the container (if enabled) and will be
                     the work directory within container.
    :param envvars: Environment variables to be used when running Ansible. Environment variables will also be
                    read from ``env/envvars`` in ``private_data_dir``
    :param passwords: A dictionary containing password prompt patterns and response values used when processing output from
                      Ansible. Passwords will also be read from ``env/passwords`` in ``private_data_dir``.
    :param settings: A dictionary containing settings values for the ``ansible-runner`` runtime environment. These will also
                     be read from ``env/settings`` in ``private_data_dir``.
    :param ssh_key: The ssh private key passed to ``ssh-agent`` as part of the ansible-playbook run.
    :param quiet: Disable all output
    :param json_mode: Store event data in place of stdout on the console and in the stdout file
    :param artifact_dir: The path to the directory where artifacts should live, this defaults to 'artifacts' under the private data dir
    :param project_dir: The path to the playbook content, this defaults to 'project' within the private data dir
    :param rotate_artifacts: Keep at most n artifact directories, disable with a value of 0 which is the default
    :param timeout: The timeout value in seconds that will be passed to either ``pexpect`` of ``subprocess`` invocation
                    (based on ``runner_mode`` selected) while executing command. It the timeout is triggered it will force cancel the
                    execution.
    :param process_isolation: Enable process isolation, using a container engine (e.g. podman).
    :param process_isolation_executable: Process isolation executable or container engine used to isolate execution. (default: podman)
    :param container_image: Container image to use when running an ansible task
    :param container_volume_mounts: List of bind mounts in the form 'host_dir:/container_dir:labels. (default: None)
    :param container_options: List of container options to pass to execution engine.
    :param container_workdir: The working directory within the container.
    :param fact_cache: A string that will be used as the name for the subdirectory of the fact cache in artifacts directory.
                       This is only used for 'jsonfile' type fact caches.
    :param fact_cache_type: A string of the type of fact cache to use.  Defaults to 'jsonfile'.
    :param private_data_dir: The directory containing all runner metadata needed to invoke the runner
                             module. Output artifacts will also be stored here for later consumption.
    :param ident: The run identifier for this invocation of Runner. Will be used to create and name
                  the artifact directory holding the results of the invocation.
    :param event_handler: An optional callback that will be invoked any time an event is received by Runner itself, return True to keep the event
    :param cancel_callback: An optional callback that can inform runner to cancel (returning True) or not (returning False)
    :param finished_callback: An optional callback that will be invoked at shutdown after process cleanup.
    :param status_handler: An optional callback that will be invoked any time the status changes (e.g...started, running, failed, successful, timeout)
    :param artifacts_handler: An optional callback that will be invoked at the end of the run to deal with the artifacts from the run.
    :param check_job_event_data: Check if job events data is completely generated. If event data is not completely generated and if
                                 value is set to 'True' it will raise 'AnsibleRunnerException' exception,
                                 if set to 'False' it log a debug message and continue execution. Default value is 'False'

    :type list_files: bool
    :type plugin_type: str
    :type response_format: str
    :type playbook_dir: str
    :type module_path: str
    :type runner_mode: str
    :type host_cwd: str
    :type envvars: dict
    :type passwords: dict
    :type settings: dict
    :type private_data_dir: str
    :type project_dir: str
    :type artifact_dir: str
    :type fact_cache_type: str
    :type fact_cache: str
    :type process_isolation: bool
    :type process_isolation_executable: str
    :type container_image: str
    :type container_volume_mounts: list
    :type container_options: list
    :type container_workdir: str
    :type ident: str
    :type rotate_artifacts: int
    :type timeout: int
    :type ssh_key: str
    :type quiet: bool
    :type json_mode: bool
    :type event_handler: Callable
    :type cancel_callback: Callable
    :type finished_callback: Callable
    :type status_handler: Callable
    :type artifacts_handler: Callable
    :type check_job_event_data: bool

    :returns: Returns a tuple of response and error string. In case if ``runner_mode`` is set to ``pexpect`` the error value is empty as
              ``pexpect`` uses same output descriptor for stdout and stderr. If the value of ``response_format`` is ``json``
              it returns a python dictionary object.
    '''
    event_callback_handler = kwargs.pop('event_handler', None)
    status_callback_handler = kwargs.pop('status_handler', None)
    artifacts_handler = kwargs.pop('artifacts_handler', None)
    cancel_callback = kwargs.pop('cancel_callback', None)
    finished_callback = kwargs.pop('finished_callback', None)

    rd = DocConfig(**kwargs)
    rd.prepare_plugin_list_command(list_files=list_files, response_format=response_format, plugin_type=plugin_type,
                                   playbook_dir=playbook_dir, module_path=module_path)
    r = Runner(rd,
               event_handler=event_callback_handler,
               status_handler=status_callback_handler,
               artifacts_handler=artifacts_handler,
               cancel_callback=cancel_callback,
               finished_callback=finished_callback)
    r.run()
    with r.stdout as stdout, r.stderr as stderr:
        response = stdout.read()
        error = stderr.read()
    if response and response_format == 'json':
        response = json.loads(sanitize_json_response(response))
    return response, error


def get_inventory(action, inventories, response_format=None, host=None, playbook_dir=None,
                  vault_ids=None, vault_password_file=None, output_file=None, export=None, **kwargs):
    '''
    Run an ansible-inventory command to get inventory related details.

    :param action: Valid values are one of ``graph``, ``host``, ``list``.
                   ``graph`` returns the inventory graph.
                   ``host`` returns info for a specific host and works as an inventory script.
                   ``list`` returns info for all hosts and also works as an inventory script.
    :param inventories: List of inventory host path.
    :param response_format: The output format for response. Valid values can be one of ``json``, ``yaml``, ``toml``.
                            Default is ``json``. If ``action`` is ``graph`` only allowed value is ``json``.
    :param host: When ``action`` is set to ``host`` this parameter is used to get the host specific information.
    :param playbook_dir: This parameter is used to sets the relative path for the inventory.
    :param vault_ids: The vault identity to use.
    :param vault_password_file: The vault password files to use.
    :param output_file: The file path in which inventory details should be sent to.
    :param export: The boolean value if set represent in a way that is optimized for export,not as an accurate
                   representation of how Ansible has processed it.
    :param runner_mode: The applicable values are ``pexpect`` and ``subprocess``. Default is set to ``subprocess``.
    :param host_cwd: The host current working directory to be mounted within the container (if enabled) and will be
                     the work directory within container.
    :param envvars: Environment variables to be used when running Ansible. Environment variables will also be
                    read from ``env/envvars`` in ``private_data_dir``
    :param passwords: A dictionary containing password prompt patterns and response values used when processing output from
                      Ansible. Passwords will also be read from ``env/passwords`` in ``private_data_dir``.
    :param settings: A dictionary containing settings values for the ``ansible-runner`` runtime environment. These will also
                     be read from ``env/settings`` in ``private_data_dir``.
    :param ssh_key: The ssh private key passed to ``ssh-agent`` as part of the ansible-playbook run.
    :param quiet: Disable all output
    :param json_mode: Store event data in place of stdout on the console and in the stdout file
    :param artifact_dir: The path to the directory where artifacts should live, this defaults to 'artifacts' under the private data dir
    :param project_dir: The path to the playbook content, this defaults to 'project' within the private data dir
    :param rotate_artifacts: Keep at most n artifact directories, disable with a value of 0 which is the default
    :param timeout: The timeout value in seconds that will be passed to either ``pexpect`` of ``subprocess`` invocation
                    (based on ``runner_mode`` selected) while executing command. It the timeout is triggered it will force cancel the
                    execution.
    :param process_isolation: Enable process isolation, using a container engine (e.g. podman).
    :param process_isolation_executable: Process isolation executable or container engine used to isolate execution. (default: podman)
    :param container_image: Container image to use when running an ansible task
    :param container_volume_mounts: List of bind mounts in the form 'host_dir:/container_dir:labels. (default: None)
    :param container_options: List of container options to pass to execution engine.
    :param container_workdir: The working directory within the container.
    :param fact_cache: A string that will be used as the name for the subdirectory of the fact cache in artifacts directory.
                       This is only used for 'jsonfile' type fact caches.
    :param fact_cache_type: A string of the type of fact cache to use.  Defaults to 'jsonfile'.
    :param private_data_dir: The directory containing all runner metadata needed to invoke the runner
                             module. Output artifacts will also be stored here for later consumption.
    :param ident: The run identifier for this invocation of Runner. Will be used to create and name
                  the artifact directory holding the results of the invocation.
    :param event_handler: An optional callback that will be invoked any time an event is received by Runner itself, return True to keep the event
    :param cancel_callback: An optional callback that can inform runner to cancel (returning True) or not (returning False)
    :param finished_callback: An optional callback that will be invoked at shutdown after process cleanup.
    :param status_handler: An optional callback that will be invoked any time the status changes (e.g...started, running, failed, successful, timeout)
    :param artifacts_handler: An optional callback that will be invoked at the end of the run to deal with the artifacts from the run.
    :param check_job_event_data: Check if job events data is completely generated. If event data is not completely generated and if
                                 value is set to 'True' it will raise 'AnsibleRunnerException' exception,
                                 if set to 'False' it log a debug message and continue execution. Default value is 'False'
    :type action: str
    :type inventories: list
    :type response_format: str
    :type host: str
    :type playbook_dir: str
    :type vault_ids: str
    :type vault_password_file: str
    :type output_file: str
    :type export: bool
    :type runner_mode: str
    :type host_cwd: str
    :type envvars: dict
    :type passwords: dict
    :type settings: dict
    :type private_data_dir: str
    :type project_dir: str
    :type artifact_dir: str
    :type fact_cache_type: str
    :type fact_cache: str
    :type process_isolation: bool
    :type process_isolation_executable: str
    :type container_image: str
    :type container_volume_mounts: list
    :type container_options: list
    :type container_workdir: str
    :type ident: str
    :type rotate_artifacts: int
    :type timeout: int
    :type ssh_key: str
    :type quiet: bool
    :type json_mode: bool
    :type event_handler: Callable
    :type cancel_callback: Callable
    :type finished_callback: Callable
    :type status_handler: Callable
    :type artifacts_handler: Callable
    :type check_job_event_data: bool

    :returns: Returns a tuple of response and error string. In case if ``runner_mode`` is set to ``pexpect`` the error value is
              empty as ``pexpect`` uses same output descriptor for stdout and stderr. If the vaue of ``response_format`` is ``json``
              it returns a python dictionary object.
    '''

    event_callback_handler = kwargs.pop('event_handler', None)
    status_callback_handler = kwargs.pop('status_handler', None)
    artifacts_handler = kwargs.pop('artifacts_handler', None)
    cancel_callback = kwargs.pop('cancel_callback', None)
    finished_callback = kwargs.pop('finished_callback', None)

    rd = InventoryConfig(**kwargs)
    rd.prepare_inventory_command(action=action, inventories=inventories, response_format=response_format, host=host, playbook_dir=playbook_dir,
                                 vault_ids=vault_ids, vault_password_file=vault_password_file, output_file=output_file, export=export)
    r = Runner(rd,
               event_handler=event_callback_handler,
               status_handler=status_callback_handler,
               artifacts_handler=artifacts_handler,
               cancel_callback=cancel_callback,
               finished_callback=finished_callback)
    r.run()
    with r.stdout as stdout, r.stderr as stderr:
        response = stdout.read()
        error = stderr.read()
    if response and response_format == 'json':
        response = json.loads(sanitize_json_response(response))
    return response, error


def get_ansible_config(action, config_file=None, only_changed=None, **kwargs):
    '''
    Run an ansible-config command to get ansible configuration releated details.

    :param action: Valid values are one of ``list``, ``dump``, ``view``
                   ``list`` returns all config options, ``dump`` returns the active configuration and
                   ``view`` returns the view of configuration file.
    :param config_file: Path to configuration file, defaults to first file found in precedence.                         .
    :param only_changed: The boolean value when set to ``True`` returns only the configurations that have changed
                         from the default. This parameter is applicable only when ``action`` is set to ``dump``.
    :param runner_mode: The applicable values are ``pexpect`` and ``subprocess``. Default is set to ``subprocess``.
    :param host_cwd: The current working directory from which the command in executable_cmd should be be executed.
    :param envvars: Environment variables to be used when running Ansible. Environment variables will also be
                    read from ``env/envvars`` in ``private_data_dir``
    :param passwords: A dictionary containing password prompt patterns and response values used when processing output from Ansible.
                      Passwords will also be read from ``env/passwords`` in ``private_data_dir``.
    :param settings: A dictionary containing settings values for the ``ansible-runner`` runtime environment. These will also
                     be read from ``env/settings`` in ``private_data_dir``.
    :param ssh_key: The ssh private key passed to ``ssh-agent`` as part of the ansible-playbook run.
    :param quiet: Disable all output
    :param json_mode: Store event data in place of stdout on the console and in the stdout file
    :param artifact_dir: The path to the directory where artifacts should live, this defaults to 'artifacts' under the private data dir
    :param project_dir: The path to the playbook content, this defaults to 'project' within the private data dir
    :param rotate_artifacts: Keep at most n artifact directories, disable with a value of 0 which is the default
    :param timeout: The timeout value in seconds that will be passed to either ``pexpect`` of ``subprocess`` invocation
                    (based on ``runner_mode`` selected) while executing command. It the timeout is triggered it will force cancel the
                    execution.
    :param process_isolation: Enable process isolation, using a container engine (e.g. podman).
    :param process_isolation_executable: Process isolation executable or container engine used to isolate execution. (default: podman)
    :param container_image: Container image to use when running an ansible task
    :param container_volume_mounts: List of bind mounts in the form 'host_dir:/container_dir:labels. (default: None)
    :param container_options: List of container options to pass to execution engine.
    :param container_workdir: The working directory within the container.
    :param fact_cache: A string that will be used as the name for the subdirectory of the fact cache in artifacts directory.
                       This is only used for 'jsonfile' type fact caches.
    :param fact_cache_type: A string of the type of fact cache to use.  Defaults to 'jsonfile'.
    :param private_data_dir: The directory containing all runner metadata needed to invoke the runner
                             module. Output artifacts will also be stored here for later consumption.
    :param ident: The run identifier for this invocation of Runner. Will be used to create and name
                  the artifact directory holding the results of the invocation.
    :param event_handler: An optional callback that will be invoked any time an event is received by Runner itself, return True to keep the event
    :param cancel_callback: An optional callback that can inform runner to cancel (returning True) or not (returning False)
    :param finished_callback: An optional callback that will be invoked at shutdown after process cleanup.
    :param status_handler: An optional callback that will be invoked any time the status changes (e.g...started, running, failed, successful, timeout)
    :param artifacts_handler: An optional callback that will be invoked at the end of the run to deal with the artifacts from the run.
    :param check_job_event_data: Check if job events data is completely generated. If event data is not completely generated and if
                                 value is set to 'True' it will raise 'AnsibleRunnerException' exception,
                                 if set to 'False' it log a debug message and continue execution. Default value is 'False'
    :type action: str
    :type config_file: str
    :type only_changed: bool
    :type runner_mode: str
    :type host_cwd: str
    :type envvars: dict
    :type passwords: dict
    :type settings: dict
    :type private_data_dir: str
    :type project_dir: str
    :type artifact_dir: str
    :type fact_cache_type: str
    :type fact_cache: str
    :type process_isolation: bool
    :type process_isolation_executable: str
    :type container_image: str
    :type container_volume_mounts: list
    :type container_options: list
    :type container_workdir: str
    :type ident: str
    :type rotate_artifacts: int
    :type timeout: int
    :type ssh_key: str
    :type quiet: bool
    :type json_mode: bool
    :type event_handler: Callable
    :type cancel_callback: Callable
    :type finished_callback: Callable
    :type status_handler: Callable
    :type artifacts_handler: Callable
    :type check_job_event_data: bool

    :returns: Returns a tuple of response and error string. In case if ``runner_mode`` is set to ``pexpect`` the error value is
              empty as ``pexpect`` uses same output descriptor for stdout and stderr.
    '''
    event_callback_handler = kwargs.pop('event_handler', None)
    status_callback_handler = kwargs.pop('status_handler', None)
    artifacts_handler = kwargs.pop('artifacts_handler', None)
    cancel_callback = kwargs.pop('cancel_callback', None)
    finished_callback = kwargs.pop('finished_callback', None)

    rd = AnsibleCfgConfig(**kwargs)
    rd.prepare_ansible_config_command(action=action, config_file=config_file, only_changed=only_changed)
    r = Runner(rd,
               event_handler=event_callback_handler,
               status_handler=status_callback_handler,
               artifacts_handler=artifacts_handler,
               cancel_callback=cancel_callback,
               finished_callback=finished_callback)
    r.run()
    with r.stdout as stdout, r.stderr as stderr:
        response = stdout.read()
        error = stderr.read()
    return response, error


def get_role_list(collection=None, playbook_dir=None, **kwargs):
    '''
    Run an ``ansible-doc`` command to get list of installed collection roles.

    Only roles that have an argument specification defined are returned.

    .. note:: Version added: 2.2

    :param str collection: A fully qualified collection name used to filter the results.
    :param str playbook_dir: This parameter is used to set the relative path to handle playbook adjacent installed roles.

    :param str runner_mode: The applicable values are ``pexpect`` and ``subprocess``. Default is set to ``subprocess``.
    :param str host_cwd: The host current working directory to be mounted within the container (if enabled) and will be
                     the work directory within container.
    :param dict envvars: Environment variables to be used when running Ansible. Environment variables will also be
                    read from ``env/envvars`` in ``private_data_dir``
    :param dict passwords: A dictionary containing password prompt patterns and response values used when processing output from
                      Ansible. Passwords will also be read from ``env/passwords`` in ``private_data_dir``.
    :param dict settings: A dictionary containing settings values for the ``ansible-runner`` runtime environment. These will also
                     be read from ``env/settings`` in ``private_data_dir``.
    :param str ssh_key: The ssh private key passed to ``ssh-agent`` as part of the ansible-playbook run.
    :param bool quiet: Disable all output
    :param bool json_mode: Store event data in place of stdout on the console and in the stdout file
    :param str artifact_dir: The path to the directory where artifacts should live, this defaults to 'artifacts' under the private data dir
    :param str project_dir: The path to the playbook content, this defaults to 'project' within the private data dir
    :param int rotate_artifacts: Keep at most n artifact directories, disable with a value of 0 which is the default
    :param int timeout: The timeout value in seconds that will be passed to either ``pexpect`` of ``subprocess`` invocation
        (based on ``runner_mode`` selected) while executing command. If the timeout is triggered, it will force cancel the execution.
    :param bool process_isolation: Enable process isolation using a container engine, such as podman.
    :param str process_isolation_executable: Process isolation executable or container engine used to isolate execution. (default: podman)
    :param str container_image: Container image to use when running an Ansible task
    :param list container_volume_mounts: List of bind mounts in the form ``host_dir:/container_dir:labels``. (default: None)
    :param list container_options: List of container options to pass to execution engine.
    :param str container_workdir: The working directory within the container.
    :param str fact_cache: A string that will be used as the name for the subdirectory of the fact cache in artifacts directory.
                       This is only used for 'jsonfile' type fact caches.
    :param str fact_cache_type: A string of the type of fact cache to use.  Defaults to 'jsonfile'.
    :param str private_data_dir: The directory containing all runner metadata needed to invoke the runner
        module. Output artifacts will also be stored here for later consumption.
    :param str ident: The run identifier for this invocation of Runner. Will be used to create and name
        the artifact directory holding the results of the invocation.
    :param Callable event_handler: An optional callback that will be invoked any time an event is received by Runner itself, return True to keep the event
    :param Callable cancel_callback: An optional callback that can inform runner to cancel (returning True) or not (returning False)
    :param Callable finished_callback: An optional callback that will be invoked at shutdown after process cleanup.
    :param Callable status_handler: An optional callback that will be invoked any time the status changes
        (for example: started, running, failed, successful, timeout)
    :param Callable artifacts_handler: An optional callback that will be invoked at the end of the run to deal with the artifacts from the run.
    :param bool check_job_event_data: Check if job events data is completely generated. If event data is not completely generated and if
        value is set to 'True' it will raise 'AnsibleRunnerException' exception. If set to 'False', log a debug message and continue execution.
        Default value is 'False'

    :returns: A tuple of response and error string. The response is a dictionary object
        (as returned by ansible-doc JSON output) containing each role found, or an empty dict
        if none are found.
    '''
    event_callback_handler = kwargs.pop('event_handler', None)
    status_callback_handler = kwargs.pop('status_handler', None)
    artifacts_handler = kwargs.pop('artifacts_handler', None)
    cancel_callback = kwargs.pop('cancel_callback', None)
    finished_callback = kwargs.pop('finished_callback', None)

    rd = DocConfig(**kwargs)
    rd.prepare_role_list_command(collection, playbook_dir)
    r = Runner(rd,
               event_handler=event_callback_handler,
               status_handler=status_callback_handler,
               artifacts_handler=artifacts_handler,
               cancel_callback=cancel_callback,
               finished_callback=finished_callback)
    r.run()
    with r.stdout as stdout, r.stderr as stderr:
        response = stdout.read()
        error = stderr.read()
    if response:
        response = json.loads(sanitize_json_response(response))
    return response, error


def get_role_argspec(role, collection=None, playbook_dir=None, **kwargs):
    '''
    Run an ``ansible-doc`` command to get a role argument specification.

    .. note:: Version added: 2.2

    :param str role: Simple role name, or fully qualified collection role name, to query.
    :param str collection: If specified, will be combined with the role name to form a fully qualified collection role name.
        If this is supplied, the ``role`` param should not be fully qualified.
    :param str playbook_dir: This parameter is used to set the relative path to handle playbook adjacent installed roles.

    :param str runner_mode: The applicable values are ``pexpect`` and ``subprocess``. Default is set to ``subprocess``.
    :param str host_cwd: The host current working directory to be mounted within the container (if enabled) and will be
                     the work directory within container.
    :param dict envvars: Environment variables to be used when running Ansible. Environment variables will also be
                    read from ``env/envvars`` in ``private_data_dir``
    :param dict passwords: A dictionary containing password prompt patterns and response values used when processing output from
                      Ansible. Passwords will also be read from ``env/passwords`` in ``private_data_dir``.
    :param dict settings: A dictionary containing settings values for the ``ansible-runner`` runtime environment. These will also
                     be read from ``env/settings`` in ``private_data_dir``.
    :param str ssh_key: The ssh private key passed to ``ssh-agent`` as part of the ansible-playbook run.
    :param bool quiet: Disable all output
    :param bool json_mode: Store event data in place of stdout on the console and in the stdout file
    :param str artifact_dir: The path to the directory where artifacts should live, this defaults to 'artifacts' under the private data dir
    :param str project_dir: The path to the playbook content, this defaults to 'project' within the private data dir
    :param int rotate_artifacts: Keep at most n artifact directories, disable with a value of 0 which is the default
    :param int timeout: The timeout value in seconds that will be passed to either ``pexpect`` of ``subprocess`` invocation
        (based on ``runner_mode`` selected) while executing command. If the timeout is triggered, it will force cancel the execution.
    :param bool process_isolation: Enable process isolation using a container engine, such as podman.
    :param str process_isolation_executable: Process isolation executable or container engine used to isolate execution. (default: podman)
    :param str container_image: Container image to use when running an Ansible task
    :param list container_volume_mounts: List of bind mounts in the form ``host_dir:/container_dir:labels``. (default: None)
    :param list container_options: List of container options to pass to execution engine.
    :param str container_workdir: The working directory within the container.
    :param str fact_cache: A string that will be used as the name for the subdirectory of the fact cache in artifacts directory.
                       This is only used for 'jsonfile' type fact caches.
    :param str fact_cache_type: A string of the type of fact cache to use.  Defaults to 'jsonfile'.
    :param str private_data_dir: The directory containing all runner metadata needed to invoke the runner
        module. Output artifacts will also be stored here for later consumption.
    :param str ident: The run identifier for this invocation of Runner. Will be used to create and name
        the artifact directory holding the results of the invocation.
    :param Callable event_handler: An optional callback that will be invoked any time an event is received by Runner itself, return True to keep the event
    :param Callable cancel_callback: An optional callback that can inform runner to cancel (returning True) or not (returning False)
    :param Callable finished_callback: An optional callback that will be invoked at shutdown after process cleanup.
    :param Callable status_handler: An optional callback that will be invoked any time the status changes
        (for example: started, running, failed, successful, timeout)
    :param Callable artifacts_handler: An optional callback that will be invoked at the end of the run to deal with the artifacts from the run.
    :param bool check_job_event_data: Check if job events data is completely generated. If event data is not completely generated and if
        value is set to 'True' it will raise 'AnsibleRunnerException' exception. If set to 'False', log a debug message and continue execution.
        Default value is 'False'

    :returns: A tuple of response and error string. The response is a dictionary object
        (as returned by ansible-doc JSON output) containing each role found, or an empty dict
        if none are found.
    '''
    event_callback_handler = kwargs.pop('event_handler', None)
    status_callback_handler = kwargs.pop('status_handler', None)
    artifacts_handler = kwargs.pop('artifacts_handler', None)
    cancel_callback = kwargs.pop('cancel_callback', None)
    finished_callback = kwargs.pop('finished_callback', None)

    rd = DocConfig(**kwargs)
    rd.prepare_role_argspec_command(role, collection, playbook_dir)
    r = Runner(rd,
               event_handler=event_callback_handler,
               status_handler=status_callback_handler,
               artifacts_handler=artifacts_handler,
               cancel_callback=cancel_callback,
               finished_callback=finished_callback)
    r.run()
    with r.stdout as stdout, r.stderr as stderr:
        response = stdout.read()
        error = stderr.read()
    if response:
        response = json.loads(sanitize_json_response(response))
    return response, error



############################
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# pylint: disable=W0201

from __future__ import annotations

import json
import logging
import os
import re
import stat
import tempfile
import shutil
from base64 import b64encode
from enum import Enum
from uuid import uuid4
from collections.abc import Mapping
from typing import Any

import pexpect

from ansible_runner import defaults
from ansible_runner.output import debug
from ansible_runner.exceptions import ConfigurationError
from ansible_runner.defaults import registry_auth_prefix
from ansible_runner.loader import ArtifactLoader
from ansible_runner.utils import (
    get_callback_dir,
    open_fifo_write,
    args2cmdline,
    sanitize_container_name,
    cli_mounts,
    register_for_cleanup,
)

logger = logging.getLogger('ansible-runner')


class BaseExecutionMode(Enum):
    NONE = 0
    # run ansible commands either locally or within EE
    ANSIBLE_COMMANDS = 1
    # execute generic commands
    GENERIC_COMMANDS = 2


class BaseConfig:

    def __init__(self,
                 private_data_dir: str | None = None,
                 host_cwd: str | None = None,
                 envvars: dict[str, Any] | None = None,
                 passwords=None,
                 settings=None,
                 project_dir: str | None = None,
                 artifact_dir: str | None = None,
                 fact_cache_type: str = 'jsonfile',
                 fact_cache=None,
                 process_isolation: bool = False,
                 process_isolation_executable: str | None = None,
                 container_image: str = "",
                 container_volume_mounts=None,
                 container_options=None,
                 container_workdir: str | None = None,
                 container_auth_data=None,
                 ident: str | None = None,
                 rotate_artifacts: int = 0,
                 timeout: int | None = None,
                 ssh_key: str | None = None,
                 quiet: bool = False,
                 json_mode: bool = False,
                 check_job_event_data: bool = False,
                 suppress_env_files: bool = False,
                 keepalive_seconds: int | None = None
                 ):
        # pylint: disable=W0613

        # common params
        self.host_cwd = host_cwd
        self.envvars = envvars
        self.ssh_key_data = ssh_key
        self.command: list[str] = []

        # container params
        self.process_isolation = process_isolation
        self.process_isolation_executable = process_isolation_executable or defaults.default_process_isolation_executable
        self.container_image = container_image
        self.container_volume_mounts = container_volume_mounts
        self.container_workdir = container_workdir
        self.container_auth_data = container_auth_data
        self.registry_auth_path: str
        self.container_name: str = ""  # like other properties, not accurate until prepare is called
        self.container_options = container_options

        # runner params
        self.rotate_artifacts = rotate_artifacts
        self.quiet = quiet
        self.json_mode = json_mode
        self.passwords = passwords
        self.settings = settings
        self.timeout = timeout
        self.check_job_event_data = check_job_event_data
        self.suppress_env_files = suppress_env_files
        # ignore this for now since it's worker-specific and would just trip up old runners
        # self.keepalive_seconds = keepalive_seconds

        # setup initial environment
        if private_data_dir:
            self.private_data_dir = os.path.abspath(private_data_dir)
            # Note that os.makedirs, exist_ok=True is dangerous.  If there's a directory writable
            # by someone other than the user anywhere in the path to be created, an attacker can
            # attempt to compromise the directories via a race.
            os.makedirs(self.private_data_dir, exist_ok=True, mode=0o700)
        else:
            self.private_data_dir = tempfile.mkdtemp(prefix=defaults.AUTO_CREATE_NAMING, dir=defaults.AUTO_CREATE_DIR)

        if artifact_dir is None:
            artifact_dir = os.path.join(self.private_data_dir, 'artifacts')
        else:
            artifact_dir = os.path.abspath(artifact_dir)

        if ident is None:
            self.ident = str(uuid4())
        else:
            self.ident = str(ident)

        self.artifact_dir = os.path.join(artifact_dir, self.ident)

        if not project_dir:
            self.project_dir = os.path.join(self.private_data_dir, 'project')
        else:
            self.project_dir = project_dir

        self.rotate_artifacts = rotate_artifacts
        self.fact_cache_type = fact_cache_type
        self.fact_cache = os.path.join(self.artifact_dir, fact_cache or 'fact_cache') if self.fact_cache_type == 'jsonfile' else None

        self.loader = ArtifactLoader(self.private_data_dir)

        if self.host_cwd:
            self.host_cwd = os.path.abspath(self.host_cwd)
            self.cwd = self.host_cwd
        else:
            self.cwd = os.getcwd()

        os.makedirs(self.artifact_dir, exist_ok=True, mode=0o700)

    _CONTAINER_ENGINES = ('docker', 'podman')

    @property
    def containerized(self):
        return self.process_isolation and self.process_isolation_executable in self._CONTAINER_ENGINES

    def prepare_env(self, runner_mode: str = 'pexpect') -> None:
        """
        Manages reading environment metadata files under ``private_data_dir`` and merging/updating
        with existing values so the :py:class:`ansible_runner.runner.Runner` object can read and use them easily
        """
        self.runner_mode = runner_mode
        try:
            if self.settings and isinstance(self.settings, dict):
                self.settings.update(self.loader.load_file('env/settings', Mapping))  # type: ignore
            else:
                self.settings = self.loader.load_file('env/settings', Mapping)
        except ConfigurationError:
            debug("Not loading settings")
            self.settings = {}

        if self.runner_mode == 'pexpect':
            try:
                if self.passwords and isinstance(self.passwords, dict):
                    self.passwords.update(self.loader.load_file('env/passwords', Mapping))  # type: ignore
                else:
                    self.passwords = self.passwords or self.loader.load_file('env/passwords', Mapping)
            except ConfigurationError:
                debug('Not loading passwords')

            self.expect_passwords = {}
            try:
                if self.passwords:
                    self.expect_passwords = {
                        re.compile(pattern, re.M): password
                        for pattern, password in self.passwords.items()
                    }
            except Exception as e:
                debug(f'Failed to compile RE from passwords: {e}')

            self.expect_passwords[pexpect.TIMEOUT] = None
            self.expect_passwords[pexpect.EOF] = None

            self.pexpect_timeout = self.settings.get('pexpect_timeout', 5)
            self.pexpect_use_poll = self.settings.get('pexpect_use_poll', True)
            self.idle_timeout = self.settings.get('idle_timeout', None)

            if self.timeout:
                self.job_timeout = int(self.timeout)
            else:
                self.job_timeout = self.settings.get('job_timeout', None)

        elif self.runner_mode == 'subprocess':
            if self.timeout:
                self.subprocess_timeout = int(self.timeout)
            else:
                self.subprocess_timeout = self.settings.get('subprocess_timeout', None)

        self.process_isolation = self.settings.get('process_isolation', self.process_isolation)
        self.process_isolation_executable = self.settings.get('process_isolation_executable', self.process_isolation_executable)

        self.container_image = self.settings.get('container_image', self.container_image)
        self.container_volume_mounts = self.settings.get('container_volume_mounts', self.container_volume_mounts)
        self.container_options = self.settings.get('container_options', self.container_options)
        self.container_auth_data = self.settings.get('container_auth_data', self.container_auth_data)

        if self.containerized:
            if not self.container_image:
                raise ConfigurationError(
                    f'container_image required when specifying process_isolation_executable={self.process_isolation_executable}'
                )
            self.container_name = f"ansible_runner_{sanitize_container_name(self.ident)}"
            self.env: dict[str, Any] = {}

            if self.process_isolation_executable == 'podman':
                # A kernel bug in RHEL < 8.5 causes podman to use the fuse-overlayfs driver. This results in errors when
                # trying to set extended file attributes. Setting this environment variable allows modules to take advantage
                # of a fallback to work around this bug when failures are encountered.
                #
                # See the following for more information:
                #    https://github.com/ansible/ansible/pull/73282
                #    https://github.com/ansible/ansible/issues/73310
                #    https://issues.redhat.com/browse/AAP-476
                self.env['ANSIBLE_UNSAFE_WRITES'] = '1'

            artifact_dir = os.path.join("/runner/artifacts", self.ident)
            self.env['AWX_ISOLATED_DATA_DIR'] = artifact_dir
            if self.fact_cache_type == 'jsonfile':
                self.env['ANSIBLE_CACHE_PLUGIN_CONNECTION'] = os.path.join(artifact_dir, 'fact_cache')

        else:
            # seed env with existing shell env
            self.env = os.environ.copy()

        if self.envvars and isinstance(self.envvars, dict):
            self.env.update(self.envvars)

        try:
            envvars = self.loader.load_file('env/envvars', Mapping)
            if envvars:
                self.env.update(envvars)  # type: ignore
        except ConfigurationError:
            debug("Not loading environment vars")
            # Still need to pass default environment to pexpect

        try:
            if self.ssh_key_data is None:
                self.ssh_key_data = self.loader.load_file('env/ssh_key', str)  # type: ignore
        except ConfigurationError:
            debug("Not loading ssh key")
            self.ssh_key_data = None

        # write the SSH key data into a fifo read by ssh-agent
        if self.ssh_key_data:
            self.ssh_key_path = os.path.join(self.artifact_dir, 'ssh_key_data')
            open_fifo_write(self.ssh_key_path, self.ssh_key_data)

        self.suppress_output_file = self.settings.get('suppress_output_file', False)
        self.suppress_ansible_output = self.settings.get('suppress_ansible_output', self.quiet)

        if 'fact_cache' in self.settings:
            if 'fact_cache_type' in self.settings:
                if self.settings['fact_cache_type'] == 'jsonfile':
                    self.fact_cache = os.path.join(self.artifact_dir, self.settings['fact_cache'])
            else:
                self.fact_cache = os.path.join(self.artifact_dir, self.settings['fact_cache'])

        # Use local callback directory
        if self.containerized:
            # when containerized, copy the callback dir to $private_data_dir/artifacts/<job_id>/callback
            # then append to env['ANSIBLE_CALLBACK_PLUGINS'] with the copied location.
            callback_dir = os.path.join(self.artifact_dir, 'callback')
            # if callback dir already exists (on repeat execution with the same ident), remove it first.
            if os.path.exists(callback_dir):
                shutil.rmtree(callback_dir)
            shutil.copytree(get_callback_dir(), callback_dir)

            container_callback_dir = os.path.join("/runner/artifacts", self.ident, "callback")
            self.env['ANSIBLE_CALLBACK_PLUGINS'] = ':'.join(filter(None, (self.env.get('ANSIBLE_CALLBACK_PLUGINS'), container_callback_dir)))
        else:
            callback_dir = self.env.get('AWX_LIB_DIRECTORY', os.getenv('AWX_LIB_DIRECTORY', ''))
            if not callback_dir:
                callback_dir = get_callback_dir()
            self.env['ANSIBLE_CALLBACK_PLUGINS'] = ':'.join(filter(None, (self.env.get('ANSIBLE_CALLBACK_PLUGINS'), callback_dir)))

        # this is an adhoc command if the module is specified, TODO: combine with logic in RunnerConfig class
        is_adhoc = bool((getattr(self, 'binary', None) is None) and (getattr(self, 'module', None) is not None))

        if self.env.get('ANSIBLE_STDOUT_CALLBACK'):
            self.env['ORIGINAL_STDOUT_CALLBACK'] = self.env.get('ANSIBLE_STDOUT_CALLBACK')

        if is_adhoc:
            # force loading awx_display stdout callback for adhoc commands
            self.env["ANSIBLE_LOAD_CALLBACK_PLUGINS"] = '1'
            if 'AD_HOC_COMMAND_ID' not in self.env:
                self.env['AD_HOC_COMMAND_ID'] = '1'

        self.env['ANSIBLE_STDOUT_CALLBACK'] = 'awx_display'

        self.env['ANSIBLE_RETRY_FILES_ENABLED'] = 'False'
        if 'ANSIBLE_HOST_KEY_CHECKING' not in self.env:
            self.env['ANSIBLE_HOST_KEY_CHECKING'] = 'False'
        if not self.containerized:
            self.env['AWX_ISOLATED_DATA_DIR'] = self.artifact_dir

        if self.fact_cache_type == 'jsonfile':
            self.env['ANSIBLE_CACHE_PLUGIN'] = 'jsonfile'
            if not self.containerized:
                self.env['ANSIBLE_CACHE_PLUGIN_CONNECTION'] = self.fact_cache

        # Pexpect will error with non-string envvars types, so we ensure string types
        self.env = {str(k): str(v) for k, v in self.env.items()}

        debug('env:')
        for k, v in sorted(self.env.items()):
            debug(f' {k}: {v}')

    def handle_command_wrap(self, execution_mode: BaseExecutionMode, cmdline_args: list[str]) -> None:
        if self.ssh_key_data:
            logger.debug('ssh key data added')
            self.command = self.wrap_args_with_ssh_agent(self.command, self.ssh_key_path)

        if self.containerized:
            logger.debug('containerization enabled')
            self.command = self.wrap_args_for_containerization(self.command, execution_mode, cmdline_args)
        else:
            logger.debug('containerization disabled')

        if hasattr(self, 'command') and isinstance(self.command, list):
            logger.debug("command: %s", ' '.join(self.command))

    def _ensure_path_safe_to_mount(self, path: str) -> None:
        if os.path.isfile(path):
            path = os.path.dirname(path)
        if os.path.join(path, "") in ('/', '/home/', '/usr/'):
            raise ConfigurationError("When using containerized execution, cannot mount '/' or '/home' or '/usr'")

    def _get_playbook_path(self, cmdline_args: list[str]) -> str | None:
        _playbook = ""
        _book_keeping_copy = cmdline_args.copy()
        for arg in cmdline_args:
            if arg in ['-i', '--inventory', '--inventory-file']:
                _book_keeping_copy_inventory_index = _book_keeping_copy.index(arg)
                _book_keeping_copy.pop(_book_keeping_copy_inventory_index)
                try:
                    _book_keeping_copy.pop(_book_keeping_copy_inventory_index)
                except IndexError:
                    # invalid command, pass through for execution
                    # to return correct error from ansible-core
                    return None

        if len(_book_keeping_copy) == 1:
            # it's probably safe to assume this is the playbook
            _playbook = _book_keeping_copy[0]
        elif _book_keeping_copy[0][0] != '-':
            # this should be the playbook, it's the only "naked" arg
            _playbook = _book_keeping_copy[0]
        else:
            # parse everything beyond the first arg because we checked that
            # in the previous case already
            for arg in _book_keeping_copy[1:]:
                if arg[0] == '-':
                    continue
                if _book_keeping_copy[(_book_keeping_copy.index(arg) - 1)][0] != '-':
                    _playbook = arg
                    break

        return _playbook

    def _update_volume_mount_paths(self,
                                   args_list: list[str],
                                   src_mount_path: str | None,
                                   dst_mount_path: str | None = None,
                                   labels: str | None = None
                                   ) -> None:

        if src_mount_path is None or not os.path.exists(src_mount_path):
            logger.debug("Source volume mount path does not exist: %s", src_mount_path)
            return

        # ensure source is abs
        src_path = os.path.abspath(os.path.expanduser(os.path.expandvars(src_mount_path)))

        # set dest src (if None) relative to workdir(not absolute) or provided
        if dst_mount_path is None:
            dst_path = src_path
        elif self.container_workdir and not os.path.isabs(dst_mount_path):
            dst_path = os.path.abspath(
                os.path.expanduser(
                    os.path.expandvars(os.path.join(self.container_workdir, dst_mount_path))
                )
            )
        else:
            dst_path = os.path.abspath(os.path.expanduser(os.path.expandvars(dst_mount_path)))

        # ensure each is a directory not file, use src for dest
        # because dest doesn't exist locally
        src_dir = src_path if os.path.isdir(src_path) else os.path.dirname(src_path)
        dst_dir = dst_path if os.path.isdir(src_path) else os.path.dirname(dst_path)

        # always ensure a trailing slash
        src_dir = os.path.join(src_dir, "")
        dst_dir = os.path.join(dst_dir, "")

        # ensure the src and dest are safe mount points
        # after stripping off the file and resolving
        self._ensure_path_safe_to_mount(src_dir)
        self._ensure_path_safe_to_mount(dst_dir)

        # format the src dest str
        volume_mount_path = f"{src_dir}:{dst_dir}"

        # add labels as needed
        if labels:
            if not labels.startswith(":"):
                volume_mount_path += ":"
            volume_mount_path += labels

        # check if mount path already added in args list
        if volume_mount_path not in args_list:
            args_list.extend(["-v", volume_mount_path])

    def _handle_ansible_cmd_options_bind_mounts(self, args_list: list[str], cmdline_args: list[str]) -> None:
        inventory_file_options = ['-i', '--inventory', '--inventory-file']
        vault_file_options = ['--vault-password-file', '--vault-pass-file']
        private_key_file_options = ['--private-key', '--key-file']

        optional_mount_args = inventory_file_options + vault_file_options + private_key_file_options

        if not cmdline_args:
            return

        if '-h' in cmdline_args or '--help' in cmdline_args:
            return

        for value in self.command:
            if 'ansible-playbook' in value:
                playbook_file_path = self._get_playbook_path(cmdline_args)
                if playbook_file_path:
                    self._update_volume_mount_paths(args_list, playbook_file_path)
                    break

        cmdline_args_copy = cmdline_args.copy()
        optional_arg_paths = []
        for arg in cmdline_args:

            if arg not in optional_mount_args:
                continue

            optional_arg_index = cmdline_args_copy.index(arg)
            optional_arg_paths.append(cmdline_args[optional_arg_index + 1])
            cmdline_args_copy.pop(optional_arg_index)
            try:
                optional_arg_value = cmdline_args_copy.pop(optional_arg_index)
            except IndexError:
                # invalid command, pass through for execution
                # to return valid error from ansible-core
                return

            if arg in inventory_file_options and optional_arg_value.endswith(','):
                # comma separated host list provided as value
                continue

            self._update_volume_mount_paths(args_list, optional_arg_value)

    def should_allocate_tty(self) -> bool:
        """Whether the container should get a ``--tty`` flag.

        The base implementation returns ``False``.
        Subclasses (e.g. :class:`~ansible_runner.config.command.CommandConfig`)
        may override this to implement richer logic.
        """
        return False

    def wrap_args_for_containerization(self,
                                       args: list[str],
                                       execution_mode: BaseExecutionMode,
                                       cmdline_args: list[str]
                                       ) -> list[str]:
        new_args = [self.process_isolation_executable]
        new_args.extend(['run', '--rm'])

        if self.runner_mode == 'pexpect' or self.should_allocate_tty():
            new_args.extend(['--tty'])

        new_args.append('--interactive')

        if self.container_workdir:
            workdir = self.container_workdir
        elif self.host_cwd is not None and os.path.exists(self.host_cwd):
            # mount current host working diretory if passed and exist
            self._ensure_path_safe_to_mount(self.host_cwd)
            self._update_volume_mount_paths(new_args, self.host_cwd)
            workdir = self.host_cwd
        else:
            workdir = "/runner/project"

        self.cwd = workdir
        new_args.extend(["--workdir", workdir])

        # For run() and run_async() API value of base execution_mode is 'BaseExecutionMode.NONE'
        # and the container volume mounts are handled separately using 'container_volume_mounts'
        # hence ignore additional mount here
        if execution_mode != BaseExecutionMode.NONE:
            if execution_mode == BaseExecutionMode.ANSIBLE_COMMANDS:
                self._handle_ansible_cmd_options_bind_mounts(new_args, cmdline_args)

            # Handle automounts for .ssh config
            self._handle_automounts(new_args)

            if 'podman' in self.process_isolation_executable:
                # container namespace stuff
                new_args.extend(["--group-add=root"])
                new_args.extend(["--ipc=host"])

            self._ensure_path_safe_to_mount(self.private_data_dir)
            # Relative paths are mounted relative to /runner/project
            for subdir in ('project', 'artifacts'):
                subdir_path = os.path.join(self.private_data_dir, subdir)
                if not os.path.exists(subdir_path):
                    os.mkdir(subdir_path, 0o700)

            # runtime commands need artifacts mounted to output data
            self._update_volume_mount_paths(new_args,
                                            f"{self.private_data_dir}/artifacts",
                                            dst_mount_path="/runner/artifacts",
                                            labels=":Z")

        else:
            subdir_path = os.path.join(self.private_data_dir, 'artifacts')
            if not os.path.exists(subdir_path):
                os.mkdir(subdir_path, 0o700)

        # Mount the entire private_data_dir
        # custom show paths inside private_data_dir do not make sense
        self._update_volume_mount_paths(new_args, self.private_data_dir, dst_mount_path="/runner", labels=":Z")

        if self.container_auth_data:
            # Pull in the necessary registry auth info, if there is a container cred
            self.registry_auth_path, registry_auth_conf_file = self._generate_container_auth_dir(self.container_auth_data)
            if 'podman' in self.process_isolation_executable:
                new_args.extend([f"--authfile={self.registry_auth_path}"])
            else:
                docker_idx = new_args.index(self.process_isolation_executable)
                new_args.insert(docker_idx + 1, f"--config={self.registry_auth_path}")
            if registry_auth_conf_file is not None:
                # Podman >= 3.1.0
                self.env['CONTAINERS_REGISTRIES_CONF'] = registry_auth_conf_file
                # Podman < 3.1.0
                self.env['REGISTRIES_CONFIG_PATH'] = registry_auth_conf_file

        if self.container_volume_mounts:
            for mapping in self.container_volume_mounts:
                volume_mounts = mapping.split(':', 2)
                self._ensure_path_safe_to_mount(volume_mounts[0])
                new_args.extend(["-v", mapping])

        # Reference the file with list of keys to pass into container
        # this file will be written in ansible_runner.runner
        env_file_host = os.path.join(self.artifact_dir, 'env.list')
        new_args.extend(['--env-file', env_file_host])

        if 'podman' in self.process_isolation_executable:
            # docker doesnt support this option
            new_args.extend(['--quiet'])

        if 'docker' in self.process_isolation_executable:
            new_args.extend([f'--user={os.getuid()}'])

        new_args.extend(['--name', self.container_name])

        if self.container_options:
            new_args.extend(self.container_options)

        new_args.extend([self.container_image])
        new_args.extend(args)
        logger.debug("container engine invocation: %s", ' '.join(new_args))
        return new_args

    def _generate_container_auth_dir(self, auth_data: dict[str, str]) -> tuple[str, str | None]:
        host = auth_data.get('host')
        token = f"{auth_data.get('username')}:{auth_data.get('password')}"
        encoded_container_auth_data = {'auths': {host: {'auth': b64encode(token.encode('UTF-8')).decode('UTF-8')}}}
        # Create a new temp file with container auth data
        path = tempfile.mkdtemp(prefix=f'{registry_auth_prefix}{self.ident}_')
        register_for_cleanup(path)

        if self.process_isolation_executable == 'docker':
            auth_filename = 'config.json'
        else:
            auth_filename = 'auth.json'
        registry_auth_path = os.path.join(path, auth_filename)
        with open(registry_auth_path, 'w') as authfile:
            os.chmod(authfile.name, stat.S_IRUSR | stat.S_IWUSR)
            authfile.write(json.dumps(encoded_container_auth_data, indent=4))

        registries_conf_path = None
        if auth_data.get('verify_ssl', True) is False:
            registries_conf_path = os.path.join(path, 'registries.conf')

            with open(registries_conf_path, 'w') as registries_conf:
                os.chmod(registries_conf.name, stat.S_IRUSR | stat.S_IWUSR)

                lines = [
                    '[[registry]]',
                    f'location = "{host}"',
                    'insecure = true',
                ]

                registries_conf.write('\n'.join(lines))

        auth_path = authfile.name
        if self.process_isolation_executable == 'docker':
            auth_path = path  # docker expects to be passed directory
        return (auth_path, registries_conf_path)

    def wrap_args_with_ssh_agent(self,
                                 args: list[str],
                                 ssh_key_path: str | None,
                                 ssh_auth_sock: str | None = None,
                                 silence_ssh_add: bool = False
                                 ) -> list[str]:
        """
        Given an existing command line and parameterization this will return the same command line wrapped with the
        necessary calls to ``ssh-agent``
        """
        if self.containerized:
            artifact_dir = os.path.join("/runner/artifacts", self.ident)
            ssh_key_path = os.path.join(artifact_dir, "ssh_key_data")

        if ssh_key_path:
            ssh_add_command = args2cmdline('ssh-add', ssh_key_path)
            if silence_ssh_add:
                ssh_add_command = ' '.join([ssh_add_command, '2>/dev/null'])
            ssh_key_cleanup_command = f'rm -f {ssh_key_path}'
            # The trap ensures the fifo is cleaned up even if the call to ssh-add fails.
            # This prevents getting into certain scenarios where subsequent reads will
            # hang forever.
            cmd = ' && '.join([args2cmdline('trap', ssh_key_cleanup_command, 'EXIT'),
                               ssh_add_command,
                               ssh_key_cleanup_command,
                               args2cmdline(*args)])
            args = ['ssh-agent']
            if ssh_auth_sock:
                args.extend(['-a', ssh_auth_sock])
            args.extend(['sh', '-c', cmd])
        return args

    def _handle_automounts(self, new_args: list[str]) -> None:
        for cli_automount in cli_mounts():
            for env in cli_automount['ENVS']:
                if env in os.environ:
                    dest_path = os.environ[env]

                    if os.path.exists(os.environ[env]):
                        if os.environ[env].startswith(os.environ['HOME']):
                            dest_path = f"/home/runner/{os.environ[env].lstrip(os.environ['HOME'])}"
                        elif os.environ[env].startswith('~'):
                            dest_path = f"/home/runner/{os.environ[env].lstrip('~/')}"
                        else:
                            dest_path = os.environ[env]

                        self._update_volume_mount_paths(new_args, os.environ[env], dst_mount_path=dest_path)

                    new_args.extend(["-e", f"{env}={dest_path}"])

            for paths in cli_automount['PATHS']:
                if os.path.exists(paths['src']):
                    self._update_volume_mount_paths(new_args, paths['src'], dst_mount_path=paths['dest'])



from __future__ import annotations

import multiprocessing
import re
import uuid

from pathlib import Path


def get_cpu_count() -> int:
    # `multiprocessing` info: https://docs.python.org/3/library/multiprocessing.html
    cpu_count = multiprocessing.cpu_count()
    return cpu_count


def get_mem_in_bytes() -> int | str:
    try:
        with open('/proc/meminfo') as f:
            mem = f.read()
        matched = re.search(r'^MemTotal:\s+(\d+)', mem)
        if matched:
            mem_capacity = int(matched.groups()[0])
        return mem_capacity * 1024
    except FileNotFoundError:
        error = "The /proc/meminfo file could not found, memory capacity undiscoverable."
        return error


def ensure_uuid(uuid_file_path: Path | None = None, mode: int = 0o0600):
    if uuid_file_path is None:
        uuid_file_path = Path.home().joinpath('.ansible_runner_uuid')

    if uuid_file_path.exists():
        uuid_file_path.chmod(mode)
        # Read the contents of file if it already exists
        saved_uuid = uuid_file_path.read_text()
        return saved_uuid.strip()

    # Generate a new UUID if file is not found
    newly_generated_uuid = _set_uuid(uuid_file_path, mode)
    return newly_generated_uuid


def _set_uuid(uuid_file_path: Path | None = None, mode: int = 0o0600):
    if uuid_file_path is None:
        uuid_file_path = Path.home().joinpath('.ansible_runner_uuid')

    generated_uuid = str(uuid.uuid4())

    if not uuid_file_path.exists():
        # Ensure the file starts with correct permissions
        uuid_file_path.touch(mode)

    # Ensure the correct permissions if the file exists
    uuid_file_path.chmod(mode)

    # Store the newly-generated UUID in a new file in home dir
    uuid_file_path.write_text(generated_uuid)

    return generated_uuid



# Copyright (c) 2016 Ansible by Red Hat, Inc.
#
# This file is part of Ansible Tower, but depends on code imported from Ansible.
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=W0212

from __future__ import (absolute_import, annotations, division, print_function)

# Python
import json
import stat
import threading
import base64
import functools
import collections
import contextlib
import datetime
import inspect
import os
import sys
import types
import typing as t
import uuid
from copy import copy

# Ansible
from ansible import __version__ as ansible_version_str
from ansible import constants as C
from ansible.plugins.callback import CallbackBase
from ansible.plugins.loader import callback_loader
from ansible.utils.display import Display

try:
    from ansible.utils.multiprocessing import context as multiprocessing_context
except ImportError:
    import multiprocessing as multiprocessing_context


DOCUMENTATION = '''
    callback: awx_display
    short_description: Playbook event dispatcher for ansible-runner
    version_added: "2.0"
    description:
        - This callback is necessary for ansible-runner to work
    type: stdout
    extends_documentation_fragment:
      - default_callback
    requirements:
      - Set as stdout in config
'''

IS_ADHOC = os.getenv('AD_HOC_COMMAND_ID') or False

# Dynamically construct base classes for our callback module, to support custom stdout callbacks.
if os.getenv('ORIGINAL_STDOUT_CALLBACK'):
    default_stdout_callback = os.getenv('ORIGINAL_STDOUT_CALLBACK')
elif IS_ADHOC:
    default_stdout_callback = 'minimal'
else:
    default_stdout_callback = 'default'

DefaultCallbackModule: CallbackBase = callback_loader.get(default_stdout_callback, class_only=True)

CENSORED = "the output has been hidden due to the fact that 'no_log: true' was specified for this result"

_ANSIBLE_VERSION = tuple(int(p) for p in ansible_version_str.split('.')[:2])
_ANSIBLE_214 = _ANSIBLE_VERSION >= (2, 14)

display = Display()


def current_time():
    return datetime.datetime.now(datetime.timezone.utc)


# use a custom JSON serializer so we can properly handle !unsafe and !vault
# objects that may exist in events emitted by the callback plugin
# see: https://github.com/ansible/ansible/pull/38759
class AnsibleJSONEncoderLocal(json.JSONEncoder):
    '''
    The class AnsibleJSONEncoder exists in Ansible core for this function
    this performs a mostly identical function via duck typing
    '''

    def default(self, o):
        '''
        Returns JSON-valid representation for special Ansible python objects
        which including vault objects and datetime objects
        '''
        if getattr(o, 'yaml_tag', None) == '!vault':
            encrypted_form = o._ciphertext
            if isinstance(encrypted_form, bytes):
                encrypted_form = encrypted_form.decode('utf-8')
            return {'__ansible_vault': encrypted_form}
        if isinstance(o, (datetime.date, datetime.datetime)):
            return o.isoformat()
        return super().default(o)


class IsolatedFileWrite:
    '''
    Class that will write partial event data to a file
    '''

    def __init__(self):
        self.private_data_dir = os.getenv('AWX_ISOLATED_DATA_DIR')

    def set(self, key, value):
        # Strip off the leading key identifying characters :1:ev-
        event_uuid = key[len(':1:ev-'):]
        # Write data in a staging area and then atomic move to pickup directory
        filename = f"{event_uuid}-partial.json"
        if not os.path.exists(os.path.join(self.private_data_dir, 'job_events')):
            os.mkdir(os.path.join(self.private_data_dir, 'job_events'), 0o700)
        dropoff_location = os.path.join(self.private_data_dir, 'job_events', filename)
        write_location = '.'.join([dropoff_location, 'tmp'])
        partial_data = json.dumps(value, cls=AnsibleJSONEncoderLocal)
        with os.fdopen(os.open(write_location, os.O_WRONLY | os.O_CREAT, stat.S_IRUSR | stat.S_IWUSR), 'w') as f:
            f.write(partial_data)
        os.rename(write_location, dropoff_location)


class EventContext:
    '''
    Store global and local (per thread/process) data associated with callback
    events and other display output methods.
    '''

    def __init__(self):
        if _ANSIBLE_214:
            self.display_lock = display._lock
        else:
            self.display_lock = multiprocessing_context.RLock()
        self._global_ctx = {}
        self._local = threading.local()
        if os.getenv('AWX_ISOLATED_DATA_DIR'):
            self.cache = IsolatedFileWrite()

    def add_local(self, **kwargs):
        tls = vars(self._local)
        ctx = tls.setdefault('_ctx', {})
        ctx.update(kwargs)

    def remove_local(self, **kwargs):
        for key in kwargs:
            self._local._ctx.pop(key, None)

    @contextlib.contextmanager
    def set_local(self, **kwargs):
        try:
            self.add_local(**kwargs)
            yield
        finally:
            self.remove_local(**kwargs)

    def get_local(self):
        return getattr(getattr(self, '_local', None), '_ctx', {})

    def add_global(self, **kwargs):
        self._global_ctx.update(kwargs)

    def remove_global(self, **kwargs):
        for key in kwargs:
            self._global_ctx.pop(key, None)

    @contextlib.contextmanager
    def set_global(self, **kwargs):
        try:
            self.add_global(**kwargs)
            yield
        finally:
            self.remove_global(**kwargs)

    def get_global(self):
        return self._global_ctx

    def get(self):
        ctx = {}
        ctx.update(self.get_global())
        ctx.update(self.get_local())
        return ctx

    def get_begin_dict(self):
        omit_event_data = os.getenv("RUNNER_OMIT_EVENTS", "False").lower() == "true"
        include_only_failed_event_data = os.getenv("RUNNER_ONLY_FAILED_EVENTS", "False").lower() == "true"
        event_data = self.get()
        event = event_data.pop('event', None)
        if not event:
            event = 'verbose'
            for key in ('debug', 'verbose', 'deprecated', 'warning', 'system_warning', 'error'):
                if event_data.get(key, False):
                    event = key
                    break
        event_dict = {'event': event}
        should_process_event_data = (include_only_failed_event_data and event in ('runner_on_failed', 'runner_on_async_failed', 'runner_on_item_failed')) \
            or not include_only_failed_event_data
        if os.getenv('JOB_ID', ''):
            event_dict['job_id'] = int(os.getenv('JOB_ID', '0'))
        if os.getenv('AD_HOC_COMMAND_ID', ''):
            event_dict['ad_hoc_command_id'] = int(os.getenv('AD_HOC_COMMAND_ID', '0'))
        if os.getenv('PROJECT_UPDATE_ID', ''):
            event_dict['project_update_id'] = int(os.getenv('PROJECT_UPDATE_ID', '0'))
        event_dict['pid'] = event_data.get('pid', os.getpid())
        event_dict['uuid'] = event_data.get('uuid', str(uuid.uuid4()))
        event_dict['created'] = event_data.get('created', current_time().isoformat())
        if not event_data.get('parent_uuid', None):
            for key in ('task_uuid', 'play_uuid', 'playbook_uuid'):
                parent_uuid = event_data.get(key, None)
                if parent_uuid and parent_uuid != event_data.get('uuid', None):
                    event_dict['parent_uuid'] = parent_uuid
                    break
        else:
            event_dict['parent_uuid'] = event_data.get('parent_uuid', None)
        if "verbosity" in event_data:
            event_dict["verbosity"] = event_data.pop("verbosity")
        if not omit_event_data and should_process_event_data:
            max_res = int(os.getenv("MAX_EVENT_RES", "700000"))
            if event not in ('playbook_on_stats',) and "res" in event_data and len(str(event_data['res'])) > max_res:
                event_data['res'] = {}
        else:
            event_data = {}
        event_dict['event_data'] = event_data
        return event_dict

    def get_end_dict(self):
        return {}

    def dump(self, fileobj, data, max_width=78, flush=False):
        b64data = base64.b64encode(json.dumps(data).encode('utf-8')).decode()
        with self.display_lock:
            # pattern corresponding to OutputEventFilter expectation
            fileobj.write('\x1b[K')
            for offset in range(0, len(b64data), max_width):
                chunk = b64data[offset:offset + max_width]
                escaped_chunk = f'{chunk}\x1b[{len(chunk)}D'
                fileobj.write(escaped_chunk)
            fileobj.write('\x1b[K')
            if flush:
                fileobj.flush()

    def dump_begin(self, fileobj):
        begin_dict = self.get_begin_dict()
        self.cache.set(f":1:ev-{begin_dict['uuid']}", begin_dict)
        self.dump(fileobj, {'uuid': begin_dict['uuid']})

    def dump_end(self, fileobj):
        self.dump(fileobj, self.get_end_dict(), flush=True)


event_context = EventContext()


@functools.cache
def _getsignature(f: t.Callable) -> inspect.Signature:
    return inspect.signature(f)


def _getcallargs(sig: inspect.Signature, *args, **kwargs) -> types.MappingProxyType:
    ba = sig.bind(*args, **kwargs)
    ba.apply_defaults()
    return types.MappingProxyType(ba.arguments)


def with_context(**context):
    global event_context  # pylint: disable=W0602

    def wrap(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            with event_context.set_local(**context):
                return f(*args, **kwargs)
        return wrapper
    return wrap


for attr in dir(Display):
    if attr.startswith('_') or 'cow' in attr or 'prompt' in attr:
        continue
    if attr in ('display', 'v', 'vv', 'vvv', 'vvvv', 'vvvvv', 'vvvvvv', 'verbose'):
        continue
    if not callable(getattr(Display, attr)):
        continue
    setattr(Display, attr, with_context(**{attr: True})(getattr(Display, attr)))


def with_verbosity(f):
    global event_context  # pylint: disable=W0602

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        sig = _getsignature(f)
        callargs = _getcallargs(sig, *args, **kwargs)
        host = callargs.get('host')
        caplevel = callargs.get('caplevel')
        context = {'verbose': True, 'verbosity': (caplevel + 1)}
        if host is not None:
            context['remote_addr'] = host
        with event_context.set_local(**context):
            return f(*args, **kwargs)
    return wrapper


Display.verbose = with_verbosity(Display.verbose)


def display_with_context(f):

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if _ANSIBLE_214 and multiprocessing_context.parent_process() is not None:
            # core 2.14 and newer proxy display, return if we are in a fork
            return f(*args, **kwargs)

        sig = _getsignature(f)
        callargs = _getcallargs(sig, *args, **kwargs)
        log_only = callargs.get('log_only')
        stderr = callargs.get('stderr')
        event_uuid = event_context.get().get('uuid', None)
        with event_context.display_lock:
            # If writing only to a log file or there is already an event UUID
            # set (from a callback module method), skip dumping the event data.
            if log_only or event_uuid:
                return f(*args, **kwargs)
            try:
                fileobj = sys.stderr if stderr else sys.stdout
                event_context.add_local(uuid=str(uuid.uuid4()))
                event_context.dump_begin(fileobj)
                return f(*args, **kwargs)
            finally:
                event_context.dump_end(fileobj)
                event_context.remove_local(uuid=None)

    return wrapper


Display.display = display_with_context(Display.display)


class CallbackModule(DefaultCallbackModule):
    '''
    Callback module for logging ansible/ansible-playbook events.
    '''

    CALLBACK_NAME = 'awx_display'

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'stdout'

    # These events should never have an associated play.
    EVENTS_WITHOUT_PLAY = [
        'playbook_on_start',
        'playbook_on_stats',
    ]

    # These events should never have an associated task.
    EVENTS_WITHOUT_TASK = EVENTS_WITHOUT_PLAY + [
        'playbook_on_setup',
        'playbook_on_notify',
        'playbook_on_import_for_host',
        'playbook_on_not_import_for_host',
        'playbook_on_no_hosts_matched',
        'playbook_on_no_hosts_remaining',
    ]

    def __init__(self):
        super().__init__()
        self._host_start = {}
        self.task_uuids = set()
        self.duplicate_task_counts = collections.defaultdict(lambda: 1)

        self.play_uuids = set()
        self.duplicate_play_counts = collections.defaultdict(lambda: 1)

        # NOTE: Ansible doesn't generate a UUID for playbook_on_start so do it for them.
        self.playbook_uuid = str(uuid.uuid4())

    def set_options(self, *args, **kwargs):
        base_config = C.config.get_configuration_definition(DefaultCallbackModule._load_name, plugin_type='callback')
        my_config = C.config.get_configuration_definition(self._load_name, plugin_type='callback')
        C.config.initialize_plugin_configuration_definitions('callback', self._load_name, base_config | my_config)
        return super().set_options(*args, **kwargs)

    @contextlib.contextmanager
    def capture_event_data(self, event, **event_data):
        event_data.setdefault('uuid', str(uuid.uuid4()))

        if event not in self.EVENTS_WITHOUT_TASK:
            task = event_data.pop('task', None)
        else:
            task = None

        if event_data.get('res'):
            if event_data['res'].get('_ansible_no_log', False):
                event_data['res'] = {'censored': CENSORED}
            if event_data['res'].get('results', []):
                event_data['res']['results'] = copy(event_data['res']['results'])
            for i, item in enumerate(event_data['res'].get('results', [])):
                if isinstance(item, dict) and item.get('_ansible_no_log', False):
                    event_data['res']['results'][i] = {'censored': CENSORED}

        with event_context.display_lock:
            try:
                event_context.add_local(event=event, **event_data)
                if task:
                    self.set_task(task, local=True)
                event_context.dump_begin(sys.stdout)
                yield
            finally:
                event_context.dump_end(sys.stdout)
                if task:
                    self.clear_task(local=True)
                event_context.remove_local(event=None, **event_data)

    def set_playbook(self, playbook):
        file_name = getattr(playbook, '_file_name', '???')
        event_context.add_global(playbook=file_name, playbook_uuid=self.playbook_uuid)
        self.clear_play()

    def set_play(self, play):
        if hasattr(play, 'hosts'):
            if isinstance(play.hosts, list):
                pattern = ','.join(play.hosts)
            else:
                pattern = play.hosts
        else:
            pattern = ''
        name = play.get_name().strip() or pattern
        event_context.add_global(play=name, play_uuid=str(play._uuid), play_pattern=pattern)
        self.clear_task()

    def clear_play(self):
        event_context.remove_global(play=None, play_uuid=None, play_pattern=None)
        self.clear_task()

    def set_task(self, task, local=False):
        self.clear_task(local)
        # FIXME: Task is "global" unless using free strategy!
        task_ctx = {
            'task': (task.name or task.action),
            'task_uuid': str(task._uuid),
            'task_action': task.action,
            'resolved_action': getattr(task, 'resolved_action', task.action),
            'task_args': '',
        }
        try:
            task_ctx['task_path'] = task.get_path()
        except AttributeError:
            pass
        if C.DISPLAY_ARGS_TO_STDOUT:  # pylint: disable=E1101
            if task.no_log:
                task_ctx['task_args'] = "the output has been hidden due to the fact that 'no_log: true' was specified for this result"
            else:
                task_args = ', '.join((f'{k}={v}' for k, v in task.args.items()))
                task_ctx['task_args'] = task_args
        if getattr(task, '_role', None):
            task_role = task._role._role_name
            if hasattr(task._role, 'get_name'):
                resolved_role = task._role.get_name()
                if resolved_role != task_role:
                    task_ctx['resolved_role'] = resolved_role
        else:
            task_role = getattr(task, 'role_name', '')
        if task_role:
            task_ctx['role'] = task_role
        if local:
            event_context.add_local(**task_ctx)
        else:
            event_context.add_global(**task_ctx)

    def clear_task(self, local=False):
        task_ctx = {
            'task': None, 'task_path': None, 'task_uuid': None, 'task_action': None, 'task_args': None, 'resolved_action': None,
            'role': None, 'resolved_role': None
        }
        if local:
            event_context.remove_local(**task_ctx)
        else:
            event_context.remove_global(**task_ctx)

    def v2_playbook_on_start(self, playbook):
        self.set_playbook(playbook)
        event_data = {
            'uuid': self.playbook_uuid,
        }
        with self.capture_event_data('playbook_on_start', **event_data):
            super().v2_playbook_on_start(playbook)

    def v2_playbook_on_vars_prompt(self, varname, private=True, prompt=None,
                                   encrypt=None, confirm=False, salt_size=None,
                                   salt=None, default=None, unsafe=None):
        event_data = {
            'varname': varname,
            'private': private,
            'prompt': prompt,
            'encrypt': encrypt,
            'confirm': confirm,
            'salt_size': salt_size,
            'salt': salt,
            'default': default,
            'unsafe': unsafe,
        }
        with self.capture_event_data('playbook_on_vars_prompt', **event_data):
            super().v2_playbook_on_vars_prompt(
                varname, private, prompt, encrypt, confirm, salt_size, salt,
                default,
            )

    def v2_playbook_on_include(self, included_file):
        event_data = {
            'included_file': included_file._filename if included_file is not None else None,
        }
        with self.capture_event_data('playbook_on_include', **event_data):
            super().v2_playbook_on_include(included_file)

    def v2_playbook_on_play_start(self, play):
        if IS_ADHOC:
            return
        play_uuid = str(play._uuid)
        if play_uuid in self.play_uuids:
            # When this play UUID repeats, it means the play is using the
            # free strategy (or serial:1) so different hosts may be running
            # different tasks within a play (where duplicate UUIDS are common).
            #
            # When this is the case, modify the UUID slightly to append
            # a counter so we can still _track_ duplicate events, but also
            # avoid breaking the display in these scenarios.
            self.duplicate_play_counts[play_uuid] += 1

            play_uuid = '_'.join([
                play_uuid,
                str(self.duplicate_play_counts[play_uuid])
            ])
        self.play_uuids.add(play_uuid)
        play._uuid = play_uuid

        self.set_play(play)
        if hasattr(play, 'hosts'):
            if isinstance(play.hosts, list):
                pattern = ','.join(play.hosts)
            else:
                pattern = play.hosts
        else:
            pattern = ''
        name = play.get_name().strip() or pattern
        event_data = {
            'name': name,
            'pattern': pattern,
            'uuid': str(play._uuid),
        }
        with self.capture_event_data('playbook_on_play_start', **event_data):
            super().v2_playbook_on_play_start(play)

    def v2_playbook_on_import_for_host(self, result, imported_file):
        # NOTE: Not used by Ansible 2.x.
        with self.capture_event_data('playbook_on_import_for_host'):
            super().v2_playbook_on_import_for_host(result, imported_file)

    def v2_playbook_on_not_import_for_host(self, result, missing_file):
        # NOTE: Not used by Ansible 2.x.
        with self.capture_event_data('playbook_on_not_import_for_host'):
            super().v2_playbook_on_not_import_for_host(result, missing_file)

    def v2_playbook_on_setup(self):
        # NOTE: Not used by Ansible 2.x.
        with self.capture_event_data('playbook_on_setup'):
            super().v2_playbook_on_setup()

    def v2_playbook_on_task_start(self, task, is_conditional):
        if IS_ADHOC:
            self.set_task(task)
            return
        # FIXME: Flag task path output as vv.
        task_uuid = str(task._uuid)
        if task_uuid in self.task_uuids:
            # When this task UUID repeats, it means the play is using the
            # free strategy (or serial:1) so different hosts may be running
            # different tasks within a play (where duplicate UUIDS are common).
            #
            # When this is the case, modify the UUID slightly to append
            # a counter so we can still _track_ duplicate events, but also
            # avoid breaking the display in these scenarios.
            self.duplicate_task_counts[task_uuid] += 1

            task_uuid = '_'.join([
                task_uuid,
                str(self.duplicate_task_counts[task_uuid])
            ])
        self.task_uuids.add(task_uuid)
        self.set_task(task)
        event_data = {
            'task': task,
            'name': task.get_name(),
            'is_conditional': is_conditional,
            'uuid': task_uuid,
        }
        with self.capture_event_data('playbook_on_task_start', **event_data):
            super().v2_playbook_on_task_start(task, is_conditional)

    def v2_playbook_on_cleanup_task_start(self, task):
        # NOTE: Not used by Ansible 2.x.
        self.set_task(task)
        event_data = {
            'task': task,
            'name': task.get_name(),
            'uuid': str(task._uuid),
            'is_conditional': True,
        }
        with self.capture_event_data('playbook_on_task_start', **event_data):
            super().v2_playbook_on_cleanup_task_start(task)

    def v2_playbook_on_handler_task_start(self, task):
        # NOTE: Re-using playbook_on_task_start event for this v2-specific
        # event, but setting is_conditional=True, which is how v1 identified a
        # task run as a handler.
        self.set_task(task)
        event_data = {
            'task': task,
            'name': task.get_name(),
            'uuid': str(task._uuid),
            'is_conditional': True,
        }
        with self.capture_event_data('playbook_on_task_start', **event_data):
            super().v2_playbook_on_handler_task_start(task)

    def v2_playbook_on_no_hosts_matched(self):
        with self.capture_event_data('playbook_on_no_hosts_matched'):
            super().v2_playbook_on_no_hosts_matched()

    def v2_playbook_on_no_hosts_remaining(self):
        with self.capture_event_data('playbook_on_no_hosts_remaining'):
            super().v2_playbook_on_no_hosts_remaining()

    def v2_playbook_on_notify(self, handler, host):
        # NOTE: Not used by Ansible < 2.5.
        event_data = {
            'host': host.get_name(),
            'handler': handler.get_name(),
        }
        with self.capture_event_data('playbook_on_notify', **event_data):
            super().v2_playbook_on_notify(handler, host)

    # ansible_stats is, retroactively, added in 2.2
    def v2_playbook_on_stats(self, stats):
        self.clear_play()
        # FIXME: Add count of plays/tasks.
        event_data = {
            'changed': stats.changed,
            'dark': stats.dark,
            'failures': stats.failures,
            'ignored': getattr(stats, 'ignored', 0),
            'ok': stats.ok,
            'processed': stats.processed,
            'rescued': getattr(stats, 'rescued', 0),
            'skipped': stats.skipped,
            'artifact_data': stats.custom.get('_run', {}) if hasattr(stats, 'custom') else {}
        }

        with self.capture_event_data('playbook_on_stats', **event_data):
            super().v2_playbook_on_stats(stats)

    @staticmethod
    def _get_event_loop(task):
        if hasattr(task, 'loop_with'):  # Ansible >=2.5
            return task.loop_with
        if hasattr(task, 'loop'):  # Ansible <2.4
            return task.loop
        return None

    def _get_result_timing_data(self, result):
        host_start = self._host_start.get(result._host.get_name())
        if host_start:
            end_time = current_time()
            return host_start, end_time, (end_time - host_start).total_seconds()
        return None, None, None

    def v2_runner_on_ok(self, result):
        # FIXME: Display detailed results or not based on verbosity.

        # strip environment vars from the job event; it already exists on the
        # job and sensitive values are filtered there
        if result._task.action in ('setup', 'gather_facts'):
            result._result.get('ansible_facts', {}).pop('ansible_env', None)

        host_start, end_time, duration = self._get_result_timing_data(result)
        event_data = {
            'host': result._host.get_name(),
            'remote_addr': result._host.address,
            'task': result._task,
            'res': result._result,
            'start': host_start,
            'end': end_time,
            'duration': duration,
            'event_loop': self._get_event_loop(result._task),
        }
        with self.capture_event_data('runner_on_ok', **event_data):
            super().v2_runner_on_ok(result)

    def v2_runner_on_failed(self, result, ignore_errors=False):
        # FIXME: Add verbosity for exception/results output.
        host_start, end_time, duration = self._get_result_timing_data(result)
        event_data = {
            'host': result._host.get_name(),
            'remote_addr': result._host.address,
            'res': result._result,
            'task': result._task,
            'start': host_start,
            'end': end_time,
            'duration': duration,
            'ignore_errors': ignore_errors,
            'event_loop': self._get_event_loop(result._task),
        }
        with self.capture_event_data('runner_on_failed', **event_data):
            super().v2_runner_on_failed(result, ignore_errors)

    def v2_runner_on_skipped(self, result):
        host_start, end_time, duration = self._get_result_timing_data(result)
        event_data = {
            'host': result._host.get_name(),
            'remote_addr': result._host.address,
            'task': result._task,
            'start': host_start,
            'end': end_time,
            'duration': duration,
            'event_loop': self._get_event_loop(result._task),
        }
        with self.capture_event_data('runner_on_skipped', **event_data):
            super().v2_runner_on_skipped(result)

    def v2_runner_on_unreachable(self, result):
        host_start, end_time, duration = self._get_result_timing_data(result)
        event_data = {
            'host': result._host.get_name(),
            'remote_addr': result._host.address,
            'task': result._task,
            'start': host_start,
            'end': end_time,
            'duration': duration,
            'res': result._result,
        }
        with self.capture_event_data('runner_on_unreachable', **event_data):
            super().v2_runner_on_unreachable(result)

    def v2_runner_on_no_hosts(self, task):
        # NOTE: Not used by Ansible 2.x.
        event_data = {
            'task': task,
        }
        with self.capture_event_data('runner_on_no_hosts', **event_data):
            super().v2_runner_on_no_hosts(task)

    def v2_runner_on_async_poll(self, result):
        # NOTE: Not used by Ansible 2.x.
        event_data = {
            'host': result._host.get_name(),
            'task': result._task,
            'res': result._result,
            'jid': result._result.get('ansible_job_id'),
        }
        with self.capture_event_data('runner_on_async_poll', **event_data):
            super().v2_runner_on_async_poll(result)

    def v2_runner_on_async_ok(self, result):
        # NOTE: Not used by Ansible 2.x.
        event_data = {
            'host': result._host.get_name(),
            'task': result._task,
            'res': result._result,
            'jid': result._result.get('ansible_job_id'),
        }
        with self.capture_event_data('runner_on_async_ok', **event_data):
            super().v2_runner_on_async_ok(result)

    def v2_runner_on_async_failed(self, result):
        # NOTE: Not used by Ansible 2.x.
        event_data = {
            'host': result._host.get_name(),
            'task': result._task,
            'res': result._result,
            'jid': result._result.get('ansible_job_id'),
        }
        with self.capture_event_data('runner_on_async_failed', **event_data):
            super().v2_runner_on_async_failed(result)

    def v2_runner_on_file_diff(self, result, diff):
        # NOTE: Not used by Ansible 2.x.
        event_data = {
            'host': result._host.get_name(),
            'task': result._task,
            'diff': diff,
        }
        with self.capture_event_data('runner_on_file_diff', **event_data):
            super().v2_runner_on_file_diff(result, diff)

    def v2_on_file_diff(self, result):
        # NOTE: Logged as runner_on_file_diff.
        event_data = {
            'host': result._host.get_name(),
            'task': result._task,
            'diff': result._result.get('diff'),
        }
        with self.capture_event_data('runner_on_file_diff', **event_data):
            super().v2_on_file_diff(result)

    def v2_runner_item_on_ok(self, result):
        event_data = {
            'host': result._host.get_name(),
            'task': result._task,
            'res': result._result,
        }
        with self.capture_event_data('runner_item_on_ok', **event_data):
            super().v2_runner_item_on_ok(result)

    def v2_runner_item_on_failed(self, result):
        event_data = {
            'host': result._host.get_name(),
            'task': result._task,
            'res': result._result,
        }
        with self.capture_event_data('runner_item_on_failed', **event_data):
            super().v2_runner_item_on_failed(result)

    def v2_runner_item_on_skipped(self, result):
        event_data = {
            'host': result._host.get_name(),
            'task': result._task,
            'res': result._result,
        }
        with self.capture_event_data('runner_item_on_skipped', **event_data):
            super().v2_runner_item_on_skipped(result)

    def v2_runner_retry(self, result):
        event_data = {
            'host': result._host.get_name(),
            'task': result._task,
            'res': result._result,
        }
        with self.capture_event_data('runner_retry', **event_data):
            super().v2_runner_retry(result)

    def v2_runner_on_start(self, host, task):
        event_data = {
            'host': host.get_name(),
            'task': task
        }
        self._host_start[host.get_name()] = current_time()
        with self.capture_event_data('runner_on_start', **event_data):
            super().v2_runner_on_start(host, task)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import annotations

import os
import json
import codecs

from typing import Any, Dict
from yaml import safe_load, YAMLError

from ansible_runner.exceptions import ConfigurationError
from ansible_runner.output import debug


class ArtifactLoader:
    '''
    Handles loading and caching file contents from disk

    This class will load the file contents and attempt to deserialize the
    contents as either JSON or YAML.  If the file contents cannot be
    deserialized, the contents will be returned to the caller as a string.

    The deserialized file contents are stored as a cached object in the
    instance to avoid any additional reads from disk for subsequent calls
    to load the same file.
    '''

    def __init__(self, base_path: str):
        self._cache: Dict[str, Any] = {}
        self.base_path = base_path

    def _load_json(self, contents: str) -> dict | None:
        '''
        Attempts to deserialize the contents of a JSON object

        :param str contents: The contents to deserialize.

        :return: A dict if the contents are JSON serialized,
            otherwise returns None.
        '''
        try:
            return json.loads(contents)
        except ValueError:
            return None

    def _load_yaml(self, contents: str) -> dict | None:
        '''
        Attempts to deserialize the contents of a YAML object.

        :param str contents: The contents to deserialize.

        :return: A dict if the contents are YAML serialized,
            otherwise returns None.
       '''
        try:
            return safe_load(contents)
        except YAMLError:
            return None

    def _get_contents(self, path: str) -> str:
        '''
        Loads the contents of the file specified by path

        :param str path: The relative or absolute path to the file to
            be loaded.  If the path is relative, then it is combined
            with the base_path to generate a full path string

        :return: The contents of the file as a string

        :raises: ConfigurationError if the file cannot be loaded.
        '''
        try:
            if not os.path.exists(path):
                raise ConfigurationError(f"specified path does not exist {path}")
            with codecs.open(path, encoding='utf-8') as f:
                data = f.read()

            return data

        except (IOError, OSError) as exc:
            raise ConfigurationError(f"error trying to load file contents: {exc}") from exc

    def abspath(self, path: str) -> str:
        '''
        Transform the path to an absolute path

        :param str path: The path to transform to an absolute path

        :return: The absolute path to the file.
        '''
        if not path.startswith(os.path.sep) or path.startswith('~'):
            path = os.path.expanduser(os.path.join(self.base_path, path))
        return path

    def isfile(self, path: str) -> bool:
        '''
        Check if the path is a file

        :param str path: The path to the file to check.  If the path is relative
            it will be exanded to an absolute path

        :return: True if path is a file, False otherwise.
        '''
        return os.path.isfile(self.abspath(path))

    def load_file(self, path: str, objtype: Any | None = None, encoding='utf-8') -> bytes | str | dict | None:
        '''
        Load the file specified by path

        This method will first try to load the file contents from cache and
        if there is a cache miss, it will load the contents from disk

        :param str path: The full or relative path to the file to be loaded.
        :param Any objtype: The object type of the file contents.  This
            is used to type check the deserialized content against the
            contents loaded from disk. Ignore serializing if objtype is str.
            Only Mapping or str types are supported.
        :param str encoding: The file contents text encoding.

        :return: The deserialized file contents which could be either a
            string object or a dict object

        :raises: ConfigurationError on error during file load or deserialization.
        '''
        parsed_data: bytes | str | dict | None
        path = self.abspath(path)
        debug(f"file path is {path}")

        if path in self._cache:
            return self._cache[path]

        try:
            debug(f"cache miss, attempting to load file from disk: {path}")
            contents = parsed_data = self._get_contents(path)
            if encoding:
                parsed_data = contents.encode(encoding)
        except ConfigurationError as exc:
            debug(str(exc))
            raise
        except UnicodeEncodeError as exc:
            raise ConfigurationError('unable to encode file contents') from exc

        if objtype is not str:
            for deserializer in (self._load_json, self._load_yaml):
                parsed_data = deserializer(contents)
                if parsed_data:
                    break

            if objtype and not isinstance(parsed_data, objtype):
                debug(f"specified file {path} is not of type {objtype}")
                raise ConfigurationError('invalid file serialization type for contents')

        self._cache[path] = parsed_data
        return parsed_data
